#!/usr/bin/env python3
# SPDX-License-Identifier: GPL-2.0
# -*- python -*-
# -*- coding: utf-8 -*-

import argparse
import perf

def main(event: str):
    evlist = perf.parse_events(event)

    for evsel in evlist:
        evsel.read_format = perf.FORMAT_TOTAL_TIME_ENABLED | perf.FORMAT_TOTAL_TIME_RUNNING

    evlist.open()
    evlist.enable()

    count = 100000
    while count > 0:
        count -= 1

    evlist.disable()

    for evsel in evlist:
        for cpu in evsel.cpus():
            for thread in evsel.threads():
                counts = evsel.read(cpu, thread)
                print(f"For {evsel} val: {counts.val} enable: {counts.ena} run: {counts.run}")

    evlist.close()

if __name__ == '__main__':
    ap = argparse.ArgumentParser()
    ap.add_argument('-e', '--event', help="Events to open", default="cpu-clock,task-clock")
    args = ap.parse_args()
    main(args.event)



from os import getenv, path
from subprocess import Popen, PIPE
from re import sub
import shlex

cc = getenv("CC")
assert cc, "Environment variable CC not set"

# Check if CC has options, as is the case in yocto, where it uses CC="cc --sysroot..."
cc_tokens = cc.split()
if len(cc_tokens) > 1:
    cc = cc_tokens[0]
    cc_options = " ".join([str(e) for e in cc_tokens[1:]]) + " "
else:
    cc_options = ""

# ignore optional stderr could be None as it is set to PIPE to avoid that.
# mypy: disable-error-code="union-attr"
cc_is_clang = b"clang version" in Popen([cc, "-v"], stderr=PIPE).stderr.readline()

srctree = getenv('srctree')
assert srctree, "Environment variable srctree, for the Linux sources, not set"
src_feature_tests  = f'{srctree}/tools/build/feature'

def clang_has_option(option):
    error_substrings = (
        b"unknown argument",
        b"is not supported",
        b"unknown warning option"
    )
    cmd = shlex.split(f"{cc} {cc_options} {option}") + [
        "-o", "/dev/null",
        path.join(src_feature_tests, "test-hello.c")
    ]
    cc_output = Popen(cmd, stderr=PIPE).stderr.readlines()
    return not any(any(error in line for error in error_substrings) for line in cc_output)

if cc_is_clang:
    from sysconfig import get_config_vars
    vars = get_config_vars()
    for var in ('CFLAGS', 'OPT'):
        vars[var] = sub("-specs=[^ ]+", "", vars[var])
        if not clang_has_option("-mcet"):
            vars[var] = sub("-mcet", "", vars[var])
        if not clang_has_option("-fcf-protection"):
            vars[var] = sub("-fcf-protection", "", vars[var])
        if not clang_has_option("-fstack-clash-protection"):
            vars[var] = sub("-fstack-clash-protection", "", vars[var])
        if not clang_has_option("-fstack-protector-strong"):
            vars[var] = sub("-fstack-protector-strong", "", vars[var])
        if not clang_has_option("-fno-semantic-interposition"):
            vars[var] = sub("-fno-semantic-interposition", "", vars[var])
        if not clang_has_option("-ffat-lto-objects"):
            vars[var] = sub("-ffat-lto-objects", "", vars[var])
        if not clang_has_option("-ftree-loop-distribute-patterns"):
            vars[var] = sub("-ftree-loop-distribute-patterns", "", vars[var])
        if not clang_has_option("-gno-variable-location-views"):
            vars[var] = sub("-gno-variable-location-views", "", vars[var])

from setuptools import setup, Extension

from setuptools.command.build_ext   import build_ext   as _build_ext
from setuptools.command.install_lib import install_lib as _install_lib

class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        self.build_lib  = build_lib
        self.build_temp = build_tmp

class install_lib(_install_lib):
    def finalize_options(self):
        _install_lib.finalize_options(self)
        self.build_dir = build_lib


# switch off several checks (need to be at the end of cflags list)
extra_cflags = ['-fno-strict-aliasing', '-Wno-write-strings', '-Wno-unused-parameter', '-Wno-redundant-decls' ]
if cc_is_clang:
    extra_cflags += ["-Wno-unused-command-line-argument" ]
    if clang_has_option("-Wno-cast-function-type-mismatch"):
        extra_cflags += ["-Wno-cast-function-type-mismatch" ]
else:
    extra_cflags += ['-Wno-cast-function-type' ]

# The python headers have mixed code with declarations (decls after asserts, for instance)
extra_cflags += [ "-Wno-declaration-after-statement" ]

src_perf  = f'{srctree}/tools/perf'
build_lib = getenv('PYTHON_EXTBUILD_LIB')
build_tmp = getenv('PYTHON_EXTBUILD_TMP')

perf = Extension('perf',
                 sources = [ src_perf + '/util/python.c' ],
		         include_dirs = ['util/include'],
		         extra_compile_args = extra_cflags,
                 )

setup(name='perf',
      version='0.1',
      description='Interface with the Linux profiling infrastructure',
      author='Arnaldo Carvalho de Melo',
      author_email='acme@redhat.com',
      license='GPLv2',
      url='http://perf.wiki.kernel.org',
      ext_modules=[perf],
      cmdclass={'build_ext': build_ext, 'install_lib': install_lib})



#!/usr/bin/env python3
# SPDX-License-Identifier: (LGPL-2.1 OR BSD-2-Clause)
"""List model names from mapfile.csv files."""
import argparse
import csv
import os
import re
from typing import List

def main() -> None:
    def dir_path(path: str) -> str:
        """Validate path is a directory for argparse."""
        if os.path.isdir(path):
            return path
        raise argparse.ArgumentTypeError(f'\'{path}\' is not a valid directory')

    def find_archs(start_dir: str, arch: str) -> List[str]:
        archs = []
        for item in os.scandir(start_dir):
            if not item.is_dir():
                continue
            if arch in (item.name, 'all'):
                archs.append(item.name)

        if len(archs) < 1:
            raise IOError(f'Missing architecture directory \'{arch}\'')

        return archs

    def find_mapfiles(start_dir: str, archs: List[str]) -> List[str]:
        result = []
        for arch in archs:
            for item in os.scandir(f'{start_dir}/{arch}'):
                if item.is_dir():
                    continue
                if item.name == 'mapfile.csv':
                    result.append(f'{start_dir}/{arch}/mapfile.csv')
        return result

    def find_cpuids(mapfiles: List[str], cpuids: str) -> List[str]:
        result = []
        for mapfile in mapfiles:
            with open(mapfile, encoding='utf-8') as csvfile:
                first = False
                table = csv.reader(csvfile)
                for row in table:
                    if not first or len(row) == 0 or row[0].startswith('#'):
                        first = True
                        continue
                    # Python regular expressions don't handle xdigit.
                    regex = row[0].replace('[[:xdigit:]]', '[0-9a-fA-F]')
                    for cpuid in cpuids.split(','):
                        if re.match(regex, cpuid):
                            result.append(row[2])
        return result

    ap = argparse.ArgumentParser()
    ap.add_argument('arch', help='Architecture name like x86')
    ap.add_argument('cpuid', default='all', help='List of cpuids to convert to model names')
    ap.add_argument(
        'starting_dir',
        type=dir_path,
        help='Root of tree containing architecture directories containing json files'
    )
    args = ap.parse_args()

    archs = find_archs(args.starting_dir, args.arch)
    mapfiles = find_mapfiles(args.starting_dir, archs)
    models = find_cpuids(mapfiles, args.cpuid)
    print(','.join(models))

if __name__ == '__main__':
    main()



#! /usr/bin/env python
# SPDX-License-Identifier: GPL-2.0
# -*- python -*-
# -*- coding: utf-8 -*-

import perf

def change_proctitle():
    try:
        import setproctitle
        setproctitle.setproctitle("tracepoint.py")
    except:
        print("Install the setproctitle python package to help with top and friends")

def main():
    change_proctitle()
    cpus    = perf.cpu_map()
    threads = perf.thread_map(-1)
    evlist = perf.parse_events("sched:sched_switch", cpus, threads)
    # Disable tracking of mmaps and similar that are unnecessary.
    for ev in evlist:
        ev.tracking = False
    # Configure evsels with default record options.
    evlist.config()
    # Simplify the sample_type and read_format of evsels
    for ev in evlist:
        ev.sample_type = ev.sample_type & ~perf.SAMPLE_IP
        ev.read_format = 0

    evlist.open()
    evlist.mmap()
    evlist.enable();

    while True:
        evlist.poll(timeout = -1)
        for cpu in cpus:
            event = evlist.read_on_cpu(cpu)
            if not event:
                continue

            if not isinstance(event, perf.sample_event):
                continue

            print("time %u prev_comm=%s prev_pid=%d prev_prio=%d prev_state=0x%x ==> next_comm=%s next_pid=%d next_prio=%d" % (
                   event.sample_time,
                   event.prev_comm,
                   event.prev_pid,
                   event.prev_prio,
                   event.prev_state,
                   event.next_comm,
                   event.next_pid,
                   event.next_prio))

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: (LGPL-2.1 OR BSD-2-Clause)
import argparse
import math
import os
from typing import Optional
from common_metrics import Cycles
from metric import (d_ratio, has_event, max, Event, JsonEncodeMetric,
                    JsonEncodeMetricGroupDescriptions, Literal, LoadEvents,
                    Metric, MetricGroup, Select)

# Global command line arguments.
_args = None
_zen_model: int = 1
interval_sec = Event("duration_time")
ins = Event("instructions")
cycles = Event("cycles")
# Number of CPU cycles scaled for SMT.
smt_cycles = Select(cycles / 2, Literal("#smt_on"), cycles)


def AmdBr():
    def Total() -> MetricGroup:
        br = Event("ex_ret_brn")
        br_m_all = Event("ex_ret_brn_misp")
        br_clr = Event("ex_ret_brn_cond_misp",
                       "ex_ret_msprd_brnch_instr_dir_msmtch",
                       "ex_ret_brn_resync")

        br_r = d_ratio(br, interval_sec)
        ins_r = d_ratio(ins, br)
        misp_r = d_ratio(br_m_all, br)
        clr_r = d_ratio(br_clr, interval_sec)

        return MetricGroup("lpm_br_total", [
            Metric("lpm_br_total_retired",
                   "The number of branch instructions retired per second.", br_r,
                   "insn/s"),
            Metric(
                "lpm_br_total_mispred",
                "The number of branch instructions retired, of any type, that were "
                "not correctly predicted as a percentage of all branch instrucions.",
                misp_r, "100%"),
            Metric("lpm_br_total_insn_between_branches",
                   "The number of instructions divided by the number of branches.",
                   ins_r, "insn"),
            Metric("lpm_br_total_insn_fe_resteers",
                   "The number of resync branches per second.", clr_r, "req/s")
        ])

    def Taken() -> MetricGroup:
        br = Event("ex_ret_brn_tkn")
        br_m_tk = Event("ex_ret_brn_tkn_misp")
        br_r = d_ratio(br, interval_sec)
        ins_r = d_ratio(ins, br)
        misp_r = d_ratio(br_m_tk, br)
        return MetricGroup("lpm_br_taken", [
            Metric("lpm_br_taken_retired",
                   "The number of taken branches that were retired per second.",
                   br_r, "insn/s"),
            Metric(
                "lpm_br_taken_mispred",
                "The number of retired taken branch instructions that were "
                "mispredicted as a percentage of all taken branches.", misp_r,
                "100%"),
            Metric(
                "lpm_br_taken_insn_between_branches",
                "The number of instructions divided by the number of taken branches.",
                ins_r, "insn"),
        ])

    def Conditional() -> Optional[MetricGroup]:
        global _zen_model
        br = Event("ex_ret_brn_cond", "ex_ret_cond")
        br_r = d_ratio(br, interval_sec)
        ins_r = d_ratio(ins, br)

        metrics = [
            Metric("lpm_br_cond_retired", "Retired conditional branch instructions.",
                   br_r, "insn/s"),
            Metric("lpm_br_cond_insn_between_branches",
                   "The number of instructions divided by the number of conditional "
                   "branches.", ins_r, "insn"),
        ]
        if _zen_model == 2:
            br_m_cond = Event("ex_ret_cond_misp")
            misp_r = d_ratio(br_m_cond, br)
            metrics += [
                Metric("lpm_br_cond_mispred",
                       "Retired conditional branch instructions mispredicted as a "
                       "percentage of all conditional branches.", misp_r, "100%"),
            ]

        return MetricGroup("lpm_br_cond", metrics)

    def Fused() -> MetricGroup:
        br = Event("ex_ret_fused_instr", "ex_ret_fus_brnch_inst")
        br_r = d_ratio(br, interval_sec)
        ins_r = d_ratio(ins, br)
        return MetricGroup("lpm_br_cond", [
            Metric("lpm_br_fused_retired",
                   "Retired fused branch instructions per second.", br_r, "insn/s"),
            Metric(
                "lpm_br_fused_insn_between_branches",
                "The number of instructions divided by the number of fused "
                "branches.", ins_r, "insn"),
        ])

    def Far() -> MetricGroup:
        br = Event("ex_ret_brn_far")
        br_r = d_ratio(br, interval_sec)
        ins_r = d_ratio(ins, br)
        return MetricGroup("lpm_br_far", [
            Metric("lpm_br_far_retired", "Retired far control transfers per second.",
                   br_r, "insn/s"),
            Metric(
                "lpm_br_far_insn_between_branches",
                "The number of instructions divided by the number of far branches.",
                ins_r, "insn"),
        ])

    return MetricGroup("lpm_br", [Total(), Taken(), Conditional(), Fused(), Far()],
                       description="breakdown of retired branch instructions")


def AmdCtxSw() -> MetricGroup:
    cs = Event("context\\-switches")
    metrics = [
        Metric("lpm_cs_rate", "Context switches per second",
               d_ratio(cs, interval_sec), "ctxsw/s")
    ]

    ev = Event("instructions")
    metrics.append(Metric("lpm_cs_instr", "Instructions per context switch",
                          d_ratio(ev, cs), "instr/cs"))

    ev = Event("cycles")
    metrics.append(Metric("lpm_cs_cycles", "Cycles per context switch",
                          d_ratio(ev, cs), "cycles/cs"))

    ev = Event("ls_dispatch.pure_ld", "ls_dispatch.ld_dispatch")
    metrics.append(Metric("lpm_cs_loads", "Loads per context switch",
                          d_ratio(ev, cs), "loads/cs"))

    ev = Event("ls_dispatch.pure_st", "ls_dispatch.store_dispatch")
    metrics.append(Metric("lpm_cs_stores", "Stores per context switch",
                          d_ratio(ev, cs), "stores/cs"))

    ev = Event("ex_ret_brn_tkn")
    metrics.append(Metric("lpm_cs_br_taken", "Branches taken per context switch",
                          d_ratio(ev, cs), "br_taken/cs"))

    return MetricGroup("lpm_cs", metrics,
                       description=("Number of context switches per second, instructions "
                                    "retired & core cycles between context switches"))


def AmdDtlb() -> Optional[MetricGroup]:
    global _zen_model
    if _zen_model >= 4:
        return None

    d_dat = Event("ls_dc_accesses") if _zen_model <= 3 else None
    d_h4k = Event("ls_l1_d_tlb_miss.tlb_reload_4k_l2_hit")
    d_hcoal = Event(
        "ls_l1_d_tlb_miss.tlb_reload_coalesced_page_hit") if _zen_model >= 2 else 0
    d_h2m = Event("ls_l1_d_tlb_miss.tlb_reload_2m_l2_hit")
    d_h1g = Event("ls_l1_d_tlb_miss.tlb_reload_1g_l2_hit")

    d_m4k = Event("ls_l1_d_tlb_miss.tlb_reload_4k_l2_miss")
    d_mcoal = Event(
        "ls_l1_d_tlb_miss.tlb_reload_coalesced_page_miss") if _zen_model >= 2 else 0
    d_m2m = Event("ls_l1_d_tlb_miss.tlb_reload_2m_l2_miss")
    d_m1g = Event("ls_l1_d_tlb_miss.tlb_reload_1g_l2_miss")

    d_w0 = Event("ls_tablewalker.dc_type0") if _zen_model <= 3 else None
    d_w1 = Event("ls_tablewalker.dc_type1") if _zen_model <= 3 else None
    walks = d_w0 + d_w1
    walks_r = d_ratio(walks, interval_sec)
    ins_w = d_ratio(ins, walks)
    l1 = d_dat
    l1_r = d_ratio(l1, interval_sec)
    l2_hits = d_h4k + d_hcoal + d_h2m + d_h1g
    l2_miss = d_m4k + d_mcoal + d_m2m + d_m1g
    l2_r = d_ratio(l2_hits + l2_miss, interval_sec)
    l1_miss = l2_hits + l2_miss + walks
    l1_hits = max(l1 - l1_miss, 0)
    ins_l = d_ratio(ins, l1_miss)

    return MetricGroup("lpm_dtlb", [
        MetricGroup("lpm_dtlb_ov", [
            Metric("lpm_dtlb_ov_insn_bt_l1_miss",
                   "DTLB overview: instructions between l1 misses.", ins_l,
                   "insns"),
            Metric("lpm_dtlb_ov_insn_bt_walks",
                   "DTLB overview: instructions between dtlb page table walks.",
                   ins_w, "insns"),
        ]),
        MetricGroup("lpm_dtlb_l1", [
            Metric("lpm_dtlb_l1_hits",
                   "DTLB L1 hits as percentage of all DTLB L1 accesses.",
                   d_ratio(l1_hits, l1), "100%"),
            Metric("lpm_dtlb_l1_miss",
                   "DTLB L1 misses as percentage of all DTLB L1 accesses.",
                   d_ratio(l1_miss, l1), "100%"),
            Metric("lpm_dtlb_l1_reqs", "DTLB L1 accesses per second.", l1_r,
                   "insns/s"),
        ]),
        MetricGroup("lpm_dtlb_l2", [
            Metric("lpm_dtlb_l2_hits",
                   "DTLB L2 hits as percentage of all DTLB L2 accesses.",
                   d_ratio(l2_hits, l2_hits + l2_miss), "100%"),
            Metric("lpm_dtlb_l2_miss",
                   "DTLB L2 misses as percentage of all DTLB L2 accesses.",
                   d_ratio(l2_miss, l2_hits + l2_miss), "100%"),
            Metric("lpm_dtlb_l2_reqs", "DTLB L2 accesses per second.", l2_r,
                   "insns/s"),
            MetricGroup("lpm_dtlb_l2_4kb", [
                Metric(
                    "lpm_dtlb_l2_4kb_hits",
                    "DTLB L2 4kb page size hits as percentage of all DTLB L2 4kb "
                    "accesses.", d_ratio(d_h4k, d_h4k + d_m4k), "100%"),
                Metric(
                    "lpm_dtlb_l2_4kb_miss",
                    "DTLB L2 4kb page size misses as percentage of all DTLB L2 4kb"
                    "accesses.", d_ratio(d_m4k, d_h4k + d_m4k), "100%")
            ]),
            MetricGroup("lpm_dtlb_l2_coalesced", [
                Metric(
                    "lpm_dtlb_l2_coal_hits",
                    "DTLB L2 coalesced page (16kb) hits as percentage of all DTLB "
                    "L2 coalesced accesses.", d_ratio(d_hcoal,
                                                      d_hcoal + d_mcoal), "100%"),
                Metric(
                    "lpm_dtlb_l2_coal_miss",
                    "DTLB L2 coalesced page (16kb) misses as percentage of all "
                    "DTLB L2 coalesced accesses.",
                    d_ratio(d_mcoal, d_hcoal + d_mcoal), "100%")
            ]),
            MetricGroup("lpm_dtlb_l2_2mb", [
                Metric(
                    "lpm_dtlb_l2_2mb_hits",
                    "DTLB L2 2mb page size hits as percentage of all DTLB L2 2mb "
                    "accesses.", d_ratio(d_h2m, d_h2m + d_m2m), "100%"),
                Metric(
                    "lpm_dtlb_l2_2mb_miss",
                    "DTLB L2 2mb page size misses as percentage of all DTLB L2 "
                    "accesses.", d_ratio(d_m2m, d_h2m + d_m2m), "100%")
            ]),
            MetricGroup("lpm_dtlb_l2_1g", [
                Metric(
                    "lpm_dtlb_l2_1g_hits",
                    "DTLB L2 1gb page size hits as percentage of all DTLB L2 1gb "
                    "accesses.", d_ratio(d_h1g, d_h1g + d_m1g), "100%"),
                Metric(
                    "lpm_dtlb_l2_1g_miss",
                    "DTLB L2 1gb page size misses as percentage of all DTLB L2 "
                    "1gb accesses.", d_ratio(d_m1g, d_h1g + d_m1g), "100%")
            ]),
        ]),
        MetricGroup("lpm_dtlb_walks", [
            Metric("lpm_dtlb_walks_reqs", "DTLB page table walks per second.",
                   walks_r, "walks/s"),
        ]),
    ], description="Data TLB metrics")


def AmdIotlb() -> Optional[MetricGroup]:
    global _zen_model
    if _zen_model < 2:
        return None

    # On AMD, the pde events cover both 2M and 1G pages.
    total_hit = Event("amd_iommu/mem_iommu_tlb_pte_hit/") + Event(
        "amd_iommu/mem_iommu_tlb_pde_hit/"
    )
    total_miss = Event("amd_iommu/mem_iommu_tlb_pte_mis/") + Event(
        "amd_iommu/mem_iommu_tlb_pde_mis/"
    )
    miss_rate = d_ratio(total_miss, total_miss + total_hit)

    interrupt_cache_hit = Event("amd_iommu/int_dte_hit/")
    interrupt_cache_miss = Event("amd_iommu/int_dte_mis/")
    interrupt_cache_lookup = interrupt_cache_hit + interrupt_cache_miss
    interrupt_cache_miss_rate = d_ratio(
        interrupt_cache_miss, interrupt_cache_miss + interrupt_cache_hit
    )

    return MetricGroup(
        "iotlb",
        [
            Metric("iotlb_total_hit", "IOTLB total hit", total_hit, "hits"),
            Metric("iotlb_total_miss", "IOTLB total miss", total_miss, "misses"),
            Metric("iotlb_miss_rate", "IOTLB miss rate", miss_rate, "100%"),
            Metric(
                "iotlb_interrupt_cache_hit",
                "IOTLB interrupt cache hit",
                interrupt_cache_hit,
                "hits",
            ),
            Metric(
                "iotlb_interrupt_cache_miss",
                "IOTLB interrupt cache miss",
                interrupt_cache_miss,
                "misses",
            ),
            Metric(
                "iotlb_interrupt_cache_lookup",
                "IOTLB interrupt cache lookup",
                interrupt_cache_lookup,
                "lookups",
            ),
            Metric(
                "iotlb_interrupt_cache_miss_rate",
                "IOTLB interrupt cache miss rate",
                interrupt_cache_miss_rate,
                "100%",
            ),
        ],
        description="IOMMU TLB metrics",
    )


def AmdItlb():
    global _zen_model
    l2h = Event("bp_l1_tlb_miss_l2_tlb_hit", "bp_l1_tlb_miss_l2_hit")
    l2m = Event("bp_l1_tlb_miss_l2_tlb_miss.all", "l2_itlb_misses",)
    l2r = l2h + l2m

    itlb_l1_mg = None
    l1m = l2r
    if _zen_model <= 3:
        l1r = Event("ic_fw32")
        l1h = max(l1r - l1m, 0)
        itlb_l1_mg = MetricGroup("lpm_itlb_l1", [
            Metric("lpm_itlb_l1_hits",
                   "L1 ITLB hits as a perecentage of L1 ITLB accesses.",
                   d_ratio(l1h, l1h + l1m), "100%"),
            Metric("lpm_itlb_l1_miss",
                   "L1 ITLB misses as a perecentage of L1 ITLB accesses.",
                   d_ratio(l1m, l1h + l1m), "100%"),
            Metric("lpm_itlb_l1_reqs",
                   "The number of 32B fetch windows transferred from IC pipe to DE "
                   "instruction decoder per second.", d_ratio(
                       l1r, interval_sec),
                   "windows/sec"),
        ])

    return MetricGroup("lpm_itlb", [
        MetricGroup("lpm_itlb_ov", [
            Metric("lpm_itlb_ov_insn_bt_l1_miss",
                   "Number of instructions between l1 misses", d_ratio(
                       ins, l1m), "insns"),
            Metric("lpm_itlb_ov_insn_bt_l2_miss",
                   "Number of instructions between l2 misses", d_ratio(
                       ins, l2m), "insns"),
        ]),
        itlb_l1_mg,
        MetricGroup("lpm_itlb_l2", [
            Metric("lpm_itlb_l2_hits",
                   "L2 ITLB hits as a percentage of all L2 ITLB accesses.",
                   d_ratio(l2h, l2r), "100%"),
            Metric("lpm_itlb_l2_miss",
                   "L2 ITLB misses as a percentage of all L2 ITLB accesses.",
                   d_ratio(l2m, l2r), "100%"),
            Metric("lpm_itlb_l2_reqs", "ITLB accesses per second.",
                   d_ratio(l2r, interval_sec), "accesses/sec"),
        ]),
    ], description="Instruction TLB breakdown")


def AmdLdSt() -> MetricGroup:
    ldst_ld = Event("ls_dispatch.pure_ld", "ls_dispatch.ld_dispatch")
    ldst_st = Event("ls_dispatch.pure_st", "ls_dispatch.store_dispatch")
    ldst_ldc1 = Event(f"{ldst_ld}/cmask=1/")
    ldst_stc1 = Event(f"{ldst_st}/cmask=1/")
    ldst_ldc2 = Event(f"{ldst_ld}/cmask=2/")
    ldst_stc2 = Event(f"{ldst_st}/cmask=2/")
    ldst_ldc3 = Event(f"{ldst_ld}/cmask=3/")
    ldst_stc3 = Event(f"{ldst_st}/cmask=3/")
    ldst_cyc = Event("ls_not_halted_cyc")

    ld_rate = d_ratio(ldst_ld, interval_sec)
    st_rate = d_ratio(ldst_st, interval_sec)

    ld_v1 = max(ldst_ldc1 - ldst_ldc2, 0)
    ld_v2 = max(ldst_ldc2 - ldst_ldc3, 0)
    ld_v3 = ldst_ldc3

    st_v1 = max(ldst_stc1 - ldst_stc2, 0)
    st_v2 = max(ldst_stc2 - ldst_stc3, 0)
    st_v3 = ldst_stc3

    return MetricGroup("lpm_ldst", [
        MetricGroup("lpm_ldst_total", [
            Metric("lpm_ldst_total_ld", "Number of loads dispatched per second.",
                   ld_rate, "insns/sec"),
            Metric("lpm_ldst_total_st", "Number of stores dispatched per second.",
                   st_rate, "insns/sec"),
        ]),
        MetricGroup("lpm_ldst_percent_insn", [
            Metric("lpm_ldst_percent_insn_ld",
                   "Load instructions as a percentage of all instructions.",
                   d_ratio(ldst_ld, ins), "100%"),
            Metric("lpm_ldst_percent_insn_st",
                   "Store instructions as a percentage of all instructions.",
                   d_ratio(ldst_st, ins), "100%"),
        ]),
        MetricGroup("lpm_ldst_ret_loads_per_cycle", [
            Metric(
                "lpm_ldst_ret_loads_per_cycle_1",
                "Load instructions retiring in 1 cycle as a percentage of all "
                "unhalted cycles.", d_ratio(ld_v1, ldst_cyc), "100%"),
            Metric(
                "lpm_ldst_ret_loads_per_cycle_2",
                "Load instructions retiring in 2 cycles as a percentage of all "
                "unhalted cycles.", d_ratio(ld_v2, ldst_cyc), "100%"),
            Metric(
                "lpm_ldst_ret_loads_per_cycle_3",
                "Load instructions retiring in 3 or more cycles as a percentage"
                "of all unhalted cycles.", d_ratio(ld_v3, ldst_cyc), "100%"),
        ]),
        MetricGroup("lpm_ldst_ret_stores_per_cycle", [
            Metric(
                "lpm_ldst_ret_stores_per_cycle_1",
                "Store instructions retiring in 1 cycle as a percentage of all "
                "unhalted cycles.", d_ratio(st_v1, ldst_cyc), "100%"),
            Metric(
                "lpm_ldst_ret_stores_per_cycle_2",
                "Store instructions retiring in 2 cycles as a percentage of all "
                "unhalted cycles.", d_ratio(st_v2, ldst_cyc), "100%"),
            Metric(
                "lpm_ldst_ret_stores_per_cycle_3",
                "Store instructions retiring in 3 or more cycles as a percentage"
                "of all unhalted cycles.", d_ratio(st_v3, ldst_cyc), "100%"),
        ]),
        MetricGroup("lpm_ldst_insn_bt", [
            Metric("lpm_ldst_insn_bt_ld", "Number of instructions between loads.",
                   d_ratio(ins, ldst_ld), "insns"),
            Metric("lpm_ldst_insn_bt_st", "Number of instructions between stores.",
                   d_ratio(ins, ldst_st), "insns"),
        ])
    ], description="Breakdown of load/store instructions")


def AmdUpc() -> Metric:
    ops = Event("ex_ret_ops", "ex_ret_cops")
    upc = d_ratio(ops, smt_cycles)
    return Metric("lpm_upc", "Micro-ops retired per core cycle (higher is better)",
                  upc, "uops/cycle")


def Idle() -> Metric:
    cyc = Event("msr/mperf/")
    tsc = Event("msr/tsc/")
    low = max(tsc - cyc, 0)
    return Metric(
        "lpm_idle",
        "Percentage of total wallclock cycles where CPUs are in low power state (C1 or deeper sleep state)",
        d_ratio(low, tsc), "100%")


def Rapl() -> MetricGroup:
    """Processor socket power consumption estimate.

    Use events from the running average power limit (RAPL) driver.
    """
    # Watts = joules/second
    # Currently only energy-pkg is supported by AMD:
    # https://lore.kernel.org/lkml/20220105185659.643355-1-eranian@google.com/
    pkg = Event("power/energy\\-pkg/")
    cond_pkg = Select(pkg, has_event(pkg), math.nan)
    scale = 2.3283064365386962890625e-10
    metrics = [
        Metric("lpm_cpu_power_pkg", "",
               d_ratio(cond_pkg * scale, interval_sec), "Watts"),
    ]

    return MetricGroup("lpm_cpu_power", metrics,
                       description="Processor socket power consumption estimates")


def UncoreL3():
    acc = Event("l3_lookup_state.all_coherent_accesses_to_l3",
                "l3_lookup_state.all_l3_req_typs")
    miss = Event("l3_lookup_state.l3_miss",
                 "l3_comb_clstr_state.request_miss")
    acc = max(acc, miss)
    hits = acc - miss

    return MetricGroup("lpm_l3", [
        Metric("lpm_l3_accesses", "L3 victim cache accesses",
               d_ratio(acc, interval_sec), "accesses/sec"),
        Metric("lpm_l3_hits", "L3 victim cache hit rate",
               d_ratio(hits, acc), "100%"),
        Metric("lpm_l3_miss", "L3 victim cache miss rate", d_ratio(miss, acc),
               "100%"),
    ], description="L3 cache breakdown per CCX")


def main() -> None:
    global _args
    global _zen_model

    def dir_path(path: str) -> str:
        """Validate path is a directory for argparse."""
        if os.path.isdir(path):
            return path
        raise argparse.ArgumentTypeError(
            f'\'{path}\' is not a valid directory')

    parser = argparse.ArgumentParser(description="AMD perf json generator")
    parser.add_argument(
        "-metricgroups", help="Generate metricgroups data", action='store_true')
    parser.add_argument("model", help="e.g. amdzen[123]")
    parser.add_argument(
        'events_path',
        type=dir_path,
        help='Root of tree containing architecture directories containing json files'
    )
    _args = parser.parse_args()

    directory = f"{_args.events_path}/x86/{_args.model}/"
    LoadEvents(directory)

    _zen_model = int(_args.model[6:])

    all_metrics = MetricGroup("", [
        AmdBr(),
        AmdCtxSw(),
        AmdDtlb(),
        AmdIotlb(),
        AmdItlb(),
        AmdLdSt(),
        AmdUpc(),
        Cycles(),
        Idle(),
        Rapl(),
        UncoreL3(),
    ])

    if _args.metricgroups:
        print(JsonEncodeMetricGroupDescriptions(all_metrics))
    else:
        print(JsonEncodeMetric(all_metrics))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: (LGPL-2.1 OR BSD-2-Clause)
import argparse
import os
from metric import (JsonEncodeMetric, JsonEncodeMetricGroupDescriptions, LoadEvents,
                    MetricGroup)
from common_metrics import Cycles

# Global command line arguments.
_args = None


def main() -> None:
    global _args

    def dir_path(path: str) -> str:
        """Validate path is a directory for argparse."""
        if os.path.isdir(path):
            return path
        raise argparse.ArgumentTypeError(
            f'\'{path}\' is not a valid directory')

    parser = argparse.ArgumentParser(description="ARM perf json generator")
    parser.add_argument(
        "-metricgroups", help="Generate metricgroups data", action='store_true')
    parser.add_argument("vendor", help="e.g. arm")
    parser.add_argument("model", help="e.g. neoverse-n1")
    parser.add_argument(
        'events_path',
        type=dir_path,
        help='Root of tree containing architecture directories containing json files'
    )
    _args = parser.parse_args()

    directory = f"{_args.events_path}/arm64/{_args.vendor}/{_args.model}/"
    LoadEvents(directory)

    all_metrics = MetricGroup("", [
        Cycles(),
    ])

    if _args.metricgroups:
        print(JsonEncodeMetricGroupDescriptions(all_metrics))
    else:
        print(JsonEncodeMetric(all_metrics))


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: (LGPL-2.1 OR BSD-2-Clause)
import json

hw_cache_id = [
    (0, # PERF_COUNT_HW_CACHE_L1D
     ["L1-dcache",  "l1-d",     "l1d",      "L1-data",],
     [0, 1, 2,], # read, write, prefetch
     "Level 1 data cache",
     ),
    (1, # PERF_COUNT_HW_CACHE_L1I
     ["L1-icache",  "l1-i",     "l1i",      "L1-instruction",],
     [0, 2,], # read, prefetch
     "Level 1 instruction cache",
     ),
    (2, # PERF_COUNT_HW_CACHE_LL
     ["LLC", "L2"],
     [0, 1, 2,], # read, write, prefetch
     "Last level cache",
     ),
    (3, # PERF_COUNT_HW_CACHE_DTLB
     ["dTLB",   "d-tlb",    "Data-TLB",],
     [0, 1, 2,], # read, write, prefetch
     "Data TLB",
     ),
    (4, # PERF_COUNT_HW_CACHE_ITLB
     ["iTLB",   "i-tlb",    "Instruction-TLB",],
     [0,], # read
     "Instruction TLB",
     ),
    (5, # PERF_COUNT_HW_CACHE_BPU
     ["branch", "branches", "bpu",      "btb",      "bpc",],
     [0,], # read
     "Branch prediction unit",
     ),
    (6, # PERF_COUNT_HW_CACHE_NODE
     ["node",],
     [0, 1, 2,], # read, write, prefetch
     "Local memory",
     ),
]

hw_cache_op = [
    (0, # PERF_COUNT_HW_CACHE_OP_READ
     ["load",   "loads",    "read",],
     "read"),
    (1, # PERF_COUNT_HW_CACHE_OP_WRITE
     ["store",  "stores",   "write",],
     "write"),
    (2, # PERF_COUNT_HW_CACHE_OP_PREFETCH
     ["prefetch",   "prefetches",   "speculative-read", "speculative-load",],
     "prefetch"),
]

hw_cache_result = [
    (0, # PERF_COUNT_HW_CACHE_RESULT_ACCESS
     ["refs",   "Reference",    "ops",      "access",],
     "accesses"),
    (1, # PERF_COUNT_HW_CACHE_RESULT_MISS
     ["misses", "miss",],
     "misses"),
]

events = []
def add_event(name: str,
              cache_id: int, cache_op: int, cache_result: int,
              desc: str,
              deprecated: bool) -> None:
    # Avoid conflicts with PERF_TYPE_HARDWARE events which are higher priority.
    if name in ["branch-misses", "branches"]:
        return

    # Tweak and deprecate L2 named events.
    if name.startswith("L2"):
        desc = desc.replace("Last level cache", "Level 2 (or higher) last level cache")
        deprecated = True

    event = {
        "EventName": name,
        "BriefDescription": desc,
        "LegacyCacheCode": f"0x{cache_id | (cache_op << 8) | (cache_result << 16):06x}",
    }

    # Deprecate events with the name starting L2 as it is actively
    # confusing as on many machines it actually means the L3 cache.
    if deprecated:
        event["Deprecated"] = "1"
    events.append(event)

for (cache_id, names, ops, cache_desc) in hw_cache_id:
    for name in names:
        add_event(name,
                  cache_id,
                  0, # PERF_COUNT_HW_CACHE_OP_READ
                  0, # PERF_COUNT_HW_CACHE_RESULT_ACCESS
                  f"{cache_desc} read accesses.",
                  deprecated=True)

        for (op, op_names, op_desc) in hw_cache_op:
            if op not in ops:
                continue
            for op_name in op_names:
                deprecated = (names[0] != name or op_names[1] != op_name)
                add_event(f"{name}-{op_name}",
                          cache_id,
                          op,
                          0, # PERF_COUNT_HW_CACHE_RESULT_ACCESS
                          f"{cache_desc} {op_desc} accesses.",
                          deprecated)

                for (result,  result_names, result_desc) in hw_cache_result:
                    for result_name in result_names:
                        deprecated = ((names[0] != name or op_names[0] != op_name) or
                                      (result == 0) or (result_names[0] != result_name))
                        add_event(f"{name}-{op_name}-{result_name}",
                                  cache_id, op, result,
                                  f"{cache_desc} {op_desc} {result_desc}.",
                                  deprecated)

        for (result,  result_names, result_desc) in hw_cache_result:
            for result_name in result_names:
                add_event(f"{name}-{result_name}",
                          cache_id,
                          0, # PERF_COUNT_HW_CACHE_OP_READ
                          result,
                          f"{cache_desc} read {result_desc}.",
                          deprecated=True)

print(json.dumps(events, indent=2))



# SPDX-License-Identifier: (LGPL-2.1 OR BSD-2-Clause)
from metric import (d_ratio, Event, Metric, MetricGroup)


def Cycles() -> MetricGroup:
    cyc_k = Event("cpu\\-cycles:kHh")  # exclude user and guest
    cyc_g = Event("cpu\\-cycles:G")   # exclude host
    cyc_u = Event("cpu\\-cycles:uH")  # exclude kernel, hypervisor and guest
    cyc = cyc_k + cyc_g + cyc_u

    return MetricGroup("lpm_cycles", [
        Metric("lpm_cycles_total", "Total number of cycles", cyc, "cycles"),
        Metric("lpm_cycles_user", "User cycles as a percentage of all cycles",
               d_ratio(cyc_u, cyc), "100%"),
        Metric("lpm_cycles_kernel", "Kernel cycles as a percentage of all cycles",
               d_ratio(cyc_k, cyc), "100%"),
        Metric("lpm_cycles_guest", "Hypervisor guest cycles as a percentage of all cycles",
               d_ratio(cyc_g, cyc), "100%"),
    ], description="cycles breakdown per privilege level (users, kernel, guest)")



#!/usr/bin/env python3
# SPDX-License-Identifier: (LGPL-2.1 OR BSD-2-Clause)
"""Interactive perf list."""

from abc import ABC, abstractmethod
import argparse
from dataclasses import dataclass
import math
from typing import Any, Dict, Optional, Tuple
import perf
from textual import on
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, HorizontalGroup, Vertical, VerticalScroll
from textual.css.query import NoMatches
from textual.command import SearchIcon
from textual.screen import ModalScreen
from textual.widgets import Button, Footer, Header, Input, Label, Sparkline, Static, Tree
from textual.widgets.tree import TreeNode


def get_info(info: Dict[str, str], key: str):
    return (info[key] + "\n") if key in info else ""


class TreeValue(ABC):
    """Abstraction for the data of value in the tree."""

    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def description(self) -> str:
        pass

    @abstractmethod
    def matches(self, query: str) -> bool:
        pass

    @abstractmethod
    def parse(self) -> perf.evlist:
        pass

    @abstractmethod
    def value(self, evlist: perf.evlist, evsel: perf.evsel, cpu: int, thread: int) -> float:
        pass


@dataclass
class Metric(TreeValue):
    """A metric in the tree."""
    metric_name: str
    metric_pmu: str

    def name(self) -> str:
        return self.metric_name

    def description(self) -> str:
        """Find and format metric description."""
        for metric in perf.metrics():
            if metric["MetricName"] != self.metric_name:
                continue
            if self.metric_pmu and metric["PMU"] != self.metric_pmu:
                continue
            desc = get_info(metric, "BriefDescription")
            desc += get_info(metric, "PublicDescription")
            desc += get_info(metric, "MetricExpr")
            desc += get_info(metric, "MetricThreshold")
            return desc
        return "description"

    def matches(self, query: str) -> bool:
        return query in self.metric_name

    def parse(self) -> perf.evlist:
        return perf.parse_metrics(self.metric_name, self.metric_pmu)

    def value(self, evlist: perf.evlist, evsel: perf.evsel, cpu: int, thread: int) -> float:
        try:
            val = evlist.compute_metric(self.metric_name, cpu, thread)
            return 0 if math.isnan(val) else val
        except:
            # Be tolerant of failures to compute metrics on particular CPUs/threads.
            return 0


@dataclass
class PmuEvent(TreeValue):
    """A PMU and event within the tree."""
    pmu: str
    event: str

    def name(self) -> str:
        if self.event.startswith(self.pmu) or ':' in self.event:
            return self.event
        else:
            return f"{self.pmu}/{self.event}/"

    def description(self) -> str:
        """Find and format event description for {pmu}/{event}/."""
        for p in perf.pmus():
            if p.name() != self.pmu:
                continue
            for info in p.events():
                if "name" not in info or info["name"] != self.event:
                    continue

                desc = get_info(info, "topic")
                desc += get_info(info, "event_type_desc")
                desc += get_info(info, "desc")
                desc += get_info(info, "long_desc")
                desc += get_info(info, "encoding_desc")
                return desc
        return "description"

    def matches(self, query: str) -> bool:
        return query in self.pmu or query in self.event

    def parse(self) -> perf.evlist:
        return perf.parse_events(self.name())

    def value(self, evlist: perf.evlist, evsel: perf.evsel, cpu: int, thread: int) -> float:
        return evsel.read(cpu, thread).val


class ErrorScreen(ModalScreen[bool]):
    """Pop up dialog for errors."""

    CSS = """
    ErrorScreen {
        align: center middle;
    }
    """

    def __init__(self, error: str):
        self.error = error
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Button(f"Error: {self.error}", variant="primary", id="error")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        self.dismiss(True)


class SearchScreen(ModalScreen[str]):
    """Pop up dialog for search."""

    CSS = """
    SearchScreen Horizontal {
        align: center middle;
        margin-top: 1;
    }
    SearchScreen Input {
        width: 1fr;
    }
    """

    def compose(self) -> ComposeResult:
        yield Horizontal(SearchIcon(), Input(placeholder="Event name"))

    def on_input_submitted(self, event: Input.Submitted) -> None:
        """Handle the user pressing Enter in the input field."""
        self.dismiss(event.value)


class Counter(HorizontalGroup):
    """Two labels for a CPU and its counter value."""

    CSS = """
    Label {
        gutter: 1;
    }
    """

    def __init__(self, cpu: int) -> None:
        self.cpu = cpu
        super().__init__()

    def compose(self) -> ComposeResult:
        label = f"cpu{self.cpu}" if self.cpu >= 0 else "total"
        yield Label(label + " ")
        yield Label("0", id=f"counter_{label}")


class CounterSparkline(HorizontalGroup):
    """A Sparkline for a performance counter."""

    def __init__(self, cpu: int) -> None:
        self.cpu = cpu
        super().__init__()

    def compose(self) -> ComposeResult:
        label = f"cpu{self.cpu}" if self.cpu >= 0 else "total"
        yield Label(label)
        yield Sparkline([], summary_function=max, id=f"sparkline_{label}")


class IListApp(App):
    TITLE = "Interactive Perf List"

    BINDINGS = [
        Binding(key="s", action="search", description="Search",
                tooltip="Search events and PMUs"),
        Binding(key="n", action="next", description="Next",
                tooltip="Next search result or item"),
        Binding(key="p", action="prev", description="Previous",
                tooltip="Previous search result or item"),
        Binding(key="c", action="collapse", description="Collapse",
                tooltip="Collapse the current PMU"),
        Binding(key="^q", action="quit", description="Quit",
                tooltip="Quit the app"),
    ]

    CSS = """
        /* Make the 'total' sparkline a different color. */
        #sparkline_total > .sparkline--min-color {
            color: $accent;
        }
        #sparkline_total > .sparkline--max-color {
            color: $accent 30%;
        }
        /*
         * Make the active_search initially not displayed with the text in
         * the middle of the line.
         */
        #active_search {
            display: none;
            width: 100%;
            text-align: center;
        }
    """

    def __init__(self, interval: float) -> None:
        self.interval = interval
        self.evlist = None
        self.selected: Optional[TreeValue] = None
        self.search_results: list[TreeNode[TreeValue]] = []
        self.cur_search_result: TreeNode[TreeValue] | None = None
        super().__init__()

    def expand_and_select(self, node: TreeNode[Any]) -> None:
        """Expand select a node in the tree."""
        if node.parent:
            node.parent.expand()
            if node.parent.parent:
                node.parent.parent.expand()
        node.expand()
        node.tree.select_node(node)
        node.tree.scroll_to_node(node)

    def set_searched_tree_node(self, previous: bool) -> None:
        """Set the cur_search_result node to either the next or previous."""
        l = len(self.search_results)

        if l < 1:
            tree: Tree[TreeValue] = self.query_one("#root", Tree)
            if previous:
                tree.action_cursor_up()
            else:
                tree.action_cursor_down()
            return

        if self.cur_search_result:
            idx = self.search_results.index(self.cur_search_result)
            if previous:
                idx = idx - 1 if idx > 0 else l - 1
            else:
                idx = idx + 1 if idx < l - 1 else 0
        else:
            idx = l - 1 if previous else 0

        node = self.search_results[idx]
        if node == self.cur_search_result:
            return

        self.cur_search_result = node
        self.expand_and_select(node)

    def action_search(self) -> None:
        """Search was chosen."""
        def set_initial_focus(event: str | None) -> None:
            """Sets the focus after the SearchScreen is dismissed."""

            search_label = self.query_one("#active_search", Label)
            search_label.display = True if event else False
            if not event:
                return
            event = event.lower()
            search_label.update(f'Searching for events matching "{event}"')

            tree: Tree[str] = self.query_one("#root", Tree)

            def find_search_results(event: str, node: TreeNode[str],
                                    cursor_seen: bool = False,
                                    match_after_cursor: Optional[TreeNode[str]] = None
                                    ) -> Tuple[bool, Optional[TreeNode[str]]]:
                """Find nodes that match the search remembering the one after the cursor."""
                if not cursor_seen and node == tree.cursor_node:
                    cursor_seen = True
                if node.data and node.data.matches(event):
                    if cursor_seen and not match_after_cursor:
                        match_after_cursor = node
                    self.search_results.append(node)

                if node.children:
                    for child in node.children:
                        (cursor_seen, match_after_cursor) = \
                            find_search_results(event, child, cursor_seen, match_after_cursor)
                return (cursor_seen, match_after_cursor)

            self.search_results.clear()
            (_, self.cur_search_result) = find_search_results(event, tree.root)
            if len(self.search_results) < 1:
                self.push_screen(ErrorScreen(f"Failed to find pmu/event or metric {event}"))
                search_label.display = False
            elif self.cur_search_result:
                self.expand_and_select(self.cur_search_result)
            else:
                self.set_searched_tree_node(previous=False)

        self.push_screen(SearchScreen(), set_initial_focus)

    def action_next(self) -> None:
        """Next was chosen."""
        self.set_searched_tree_node(previous=False)

    def action_prev(self) -> None:
        """Previous was chosen."""
        self.set_searched_tree_node(previous=True)

    def action_collapse(self) -> None:
        """Collapse the part of the tree currently on."""
        tree: Tree[str] = self.query_one("#root", Tree)
        node = tree.cursor_node
        if node and node.parent:
            node.parent.collapse_all()
            node.tree.scroll_to_node(node.parent)

    def update_counts(self) -> None:
        """Called every interval to update counts."""
        if not self.selected or not self.evlist:
            return

        def update_count(cpu: int, count: int):
            # Update the raw count display.
            counter: Label = self.query(f"#counter_cpu{cpu}" if cpu >= 0 else "#counter_total")
            if not counter:
                return
            counter = counter.first(Label)
            counter.update(str(count))

            # Update the sparkline.
            line: Sparkline = self.query(f"#sparkline_cpu{cpu}" if cpu >= 0 else "#sparkline_total")
            if not line:
                return
            line = line.first(Sparkline)
            # If there are more events than the width, remove the front event.
            if len(line.data) > line.size.width:
                line.data.pop(0)
            line.data.append(count)
            line.mutate_reactive(Sparkline.data)

        # Update the total and each CPU counts, assume there's just 1 evsel.
        total = 0
        self.evlist.disable()
        for evsel in self.evlist:
            for cpu in evsel.cpus():
                aggr = 0
                for thread in evsel.threads():
                    aggr += self.selected.value(self.evlist, evsel, cpu, thread)
                update_count(cpu, aggr)
                total += aggr
        update_count(-1, total)
        self.evlist.enable()

    def on_mount(self) -> None:
        """When App starts set up periodic event updating."""
        self.update_counts()
        self.set_interval(self.interval, self.update_counts)

    def set_selected(self, value: TreeValue) -> None:
        """Updates the event/description and starts the counters."""
        try:
            label_name = self.query_one("#event_name", Label)
            event_description = self.query_one("#event_description", Static)
            lines = self.query_one("#lines")
        except NoMatches:
            # A race with rendering, ignore the update as we can't
            # mount the assumed output widgets.
            return

        self.selected = value

        # Remove previous event information.
        if self.evlist:
            self.evlist.disable()
            self.evlist.close()
            old_lines = self.query(CounterSparkline)
            for line in old_lines:
                line.remove()
            old_counters = self.query(Counter)
            for counter in old_counters:
                counter.remove()

        # Update event/metric text and description.
        label_name.update(value.name())
        event_description.update(value.description())

        # Open the event.
        try:
            self.evlist = value.parse()
            if self.evlist:
                self.evlist.open()
                self.evlist.enable()
        except:
            self.evlist = None

        if not self.evlist:
            self.push_screen(ErrorScreen(f"Failed to open {value.name()}"))
            return

        # Add spark lines for all the CPUs. Note, must be done after
        # open so that the evlist CPUs have been computed by propagate
        # maps.
        line = CounterSparkline(cpu=-1)
        lines.mount(line)
        for cpu in self.evlist.all_cpus():
            line = CounterSparkline(cpu)
            lines.mount(line)
        line = Counter(cpu=-1)
        lines.mount(line)
        for cpu in self.evlist.all_cpus():
            line = Counter(cpu)
            lines.mount(line)

    def compose(self) -> ComposeResult:
        """Draws the app."""
        def metric_event_tree() -> Tree:
            """Create tree of PMUs and metricgroups with events or metrics under."""
            tree: Tree[TreeValue] = Tree("Root", id="root")
            pmus = tree.root.add("PMUs")
            for pmu in perf.pmus():
                pmu_name = pmu.name().lower()
                pmu_node = pmus.add(pmu_name)
                try:
                    for event in sorted(pmu.events(), key=lambda x: x["name"]):
                        if "deprecated" in event:
                            continue
                        if "name" in event:
                            e = event["name"].lower()
                            if "alias" in event:
                                pmu_node.add_leaf(f'{e} ({event["alias"]})',
                                                  data=PmuEvent(pmu_name, e))
                            else:
                                pmu_node.add_leaf(e, data=PmuEvent(pmu_name, e))
                except:
                    # Reading events may fail with EPERM, ignore.
                    pass
            metrics = tree.root.add("Metrics")
            groups = set()
            for metric in perf.metrics():
                groups.update(metric["MetricGroup"])

            def add_metrics_to_tree(node: TreeNode[TreeValue], parent: str, pmu: str = None):
                for metric in sorted(perf.metrics(), key=lambda x: x["MetricName"]):
                    metric_pmu = metric.get('PMU')
                    if pmu and metric_pmu and metric_pmu != pmu:
                        continue
                    if parent in metric["MetricGroup"]:
                        name = metric["MetricName"]
                        display_name = name
                        if metric_pmu:
                            display_name += f" ({metric_pmu})"
                        node.add_leaf(display_name, data=Metric(name, metric_pmu))
                        child_group_name = f'{name}_group'
                        if child_group_name in groups:
                            display_child_group_name = child_group_name
                            if metric_pmu:
                                display_child_group_name += f" ({metric_pmu})"
                            add_metrics_to_tree(node.add(display_child_group_name),
                                                child_group_name,
                                                metric_pmu)

            for group in sorted(groups):
                if group.endswith('_group'):
                    continue
                add_metrics_to_tree(metrics.add(group), group)

            tree.root.expand()
            return tree

        yield Header(id="header")
        yield Horizontal(Vertical(metric_event_tree(), id="events"),
                         Vertical(Label("event name", id="event_name"),
                                  Static("description", markup=False, id="event_description"),
                                  ))
        yield Label(id="active_search")
        yield VerticalScroll(id="lines")
        yield Footer(id="footer")

    @on(Tree.NodeSelected)
    def on_tree_node_selected(self, event: Tree.NodeSelected[TreeValue]) -> None:
        """Called when a tree node is selected, selecting the event."""
        if event.node.data:
            self.set_selected(event.node.data)


if __name__ == "__main__":
    ap = argparse.ArgumentParser()
    ap.add_argument('-I', '--interval', help="Counter update interval in seconds", default=0.1)
    args = ap.parse_args()
    app = IListApp(float(args.interval))
    app.run()



#!/usr/bin/env python3
# SPDX-License-Identifier: (LGPL-2.1 OR BSD-2-Clause)
import argparse
import json
import math
import os
import re
from typing import Optional
from common_metrics import Cycles
from metric import (d_ratio, has_event, max, aggr_nr, CheckPmu, Event,
                    JsonEncodeMetric, JsonEncodeMetricGroupDescriptions,
                    Literal, LoadEvents, Metric, MetricConstraint, MetricGroup,
                    MetricRef, Select)

# Global command line arguments.
_args = None
interval_sec = Event("duration_time")


def Idle() -> Metric:
    cyc = Event("msr/mperf/")
    tsc = Event("msr/tsc/")
    low = max(tsc - cyc, 0)
    return Metric(
        "lpm_idle",
        "Percentage of total wallclock cycles where CPUs are in low power state (C1 or deeper sleep state)",
        d_ratio(low, tsc), "100%")


def Rapl() -> MetricGroup:
    """Processor power consumption estimate.

    Use events from the running average power limit (RAPL) driver.
    """
    # Watts = joules/second
    pkg = Event("power/energy\\-pkg/")
    cond_pkg = Select(pkg, has_event(pkg), math.nan)
    cores = Event("power/energy\\-cores/")
    cond_cores = Select(cores, has_event(cores), math.nan)
    ram = Event("power/energy\\-ram/")
    cond_ram = Select(ram, has_event(ram), math.nan)
    gpu = Event("power/energy\\-gpu/")
    cond_gpu = Select(gpu, has_event(gpu), math.nan)
    psys = Event("power/energy\\-psys/")
    cond_psys = Select(psys, has_event(psys), math.nan)
    scale = 2.3283064365386962890625e-10
    metrics = [
        Metric("lpm_cpu_power_pkg", "",
               d_ratio(cond_pkg * scale, interval_sec), "Watts"),
        Metric("lpm_cpu_power_cores", "",
               d_ratio(cond_cores * scale, interval_sec), "Watts"),
        Metric("lpm_cpu_power_ram", "",
               d_ratio(cond_ram * scale, interval_sec), "Watts"),
        Metric("lpm_cpu_power_gpu", "",
               d_ratio(cond_gpu * scale, interval_sec), "Watts"),
        Metric("lpm_cpu_power_psys", "",
               d_ratio(cond_psys * scale, interval_sec), "Watts"),
    ]

    return MetricGroup("lpm_cpu_power", metrics,
                       description="Running Average Power Limit (RAPL) power consumption estimates")


def Smi() -> MetricGroup:
    pmu = "<cpu_core or cpu_atom>" if CheckPmu("cpu_core") else "cpu"
    aperf = Event('msr/aperf/')
    cycles = Event('cycles')
    smi_num = Event('msr/smi/')
    smi_cycles = Select(Select((aperf - cycles) / aperf, smi_num > 0, 0),
                        has_event(aperf),
                        0)
    return MetricGroup('smi', [
        Metric('smi_num', 'Number of SMI interrupts.',
               Select(smi_num, has_event(smi_num), 0), 'SMI#'),
        # Note, the smi_cycles "Event" is really a reference to the metric.
        Metric('smi_cycles',
               'Percentage of cycles spent in System Management Interrupts. '
               f'Requires /sys/bus/event_source/devices/{pmu}/freeze_on_smi to be 1.',
               smi_cycles, '100%', threshold=(MetricRef('smi_cycles') > 0.10))
    ], description='System Management Interrupt metrics')


def Tsx() -> Optional[MetricGroup]:
    pmu = "cpu_core" if CheckPmu("cpu_core") else "cpu"
    cycles = Event('cycles')
    cycles_in_tx = Event(f'{pmu}/cycles\\-t/')
    cycles_in_tx_cp = Event(f'{pmu}/cycles\\-ct/')
    try:
        # Test if the tsx event is present in the json, prefer the
        # sysfs version so that we can detect its presence at runtime.
        transaction_start = Event("RTM_RETIRED.START")
        transaction_start = Event(f'{pmu}/tx\\-start/')
    except:
        return None

    elision_start = None
    try:
        # Elision start isn't supported by all models, but we'll not
        # generate the tsx_cycles_per_elision metric in that
        # case. Again, prefer the sysfs encoding of the event.
        elision_start = Event("HLE_RETIRED.START")
        elision_start = Event(f'{pmu}/el\\-start/')
    except:
        pass

    return MetricGroup('transaction', [
        Metric('tsx_transactional_cycles',
               'Percentage of cycles within a transaction region.',
               Select(cycles_in_tx / cycles, has_event(cycles_in_tx), 0),
               '100%'),
        Metric('tsx_aborted_cycles', 'Percentage of cycles in aborted transactions.',
               Select(max(cycles_in_tx - cycles_in_tx_cp, 0) / cycles,
                      has_event(cycles_in_tx),
                      0),
               '100%'),
        Metric('tsx_cycles_per_transaction',
               'Number of cycles within a transaction divided by the number of transactions.',
               Select(cycles_in_tx / transaction_start,
                      has_event(cycles_in_tx),
                      0),
               "cycles / transaction"),
        Metric('tsx_cycles_per_elision',
               'Number of cycles within a transaction divided by the number of elisions.',
               Select(cycles_in_tx / elision_start,
                      has_event(elision_start),
                      0),
               "cycles / elision") if elision_start else None,
    ], description="Breakdown of transactional memory statistics")


def IntelBr():
    ins = Event("instructions")

    def Total() -> MetricGroup:
        br_all = Event("BR_INST_RETIRED.ALL_BRANCHES", "BR_INST_RETIRED.ANY")
        br_m_all = Event("BR_MISP_RETIRED.ALL_BRANCHES",
                         "BR_INST_RETIRED.MISPRED",
                         "BR_MISP_EXEC.ANY")
        br_clr = None
        try:
            br_clr = Event("BACLEARS.ANY", "BACLEARS.ALL")
        except:
            pass

        br_r = d_ratio(br_all, interval_sec)
        ins_r = d_ratio(ins, br_all)
        misp_r = d_ratio(br_m_all, br_all)
        clr_r = d_ratio(br_clr, interval_sec) if br_clr else None

        return MetricGroup("lpm_br_total", [
            Metric("lpm_br_total_retired",
                   "The number of branch instructions retired per second.", br_r,
                   "insn/s"),
            Metric(
                "lpm_br_total_mispred",
                "The number of branch instructions retired, of any type, that were "
                "not correctly predicted as a percentage of all branch instrucions.",
                misp_r, "100%"),
            Metric("lpm_br_total_insn_between_branches",
                   "The number of instructions divided by the number of branches.",
                   ins_r, "insn"),
            Metric("lpm_br_total_insn_fe_resteers",
                   "The number of resync branches per second.", clr_r, "req/s"
                   ) if clr_r else None
        ])

    def Taken() -> MetricGroup:
        br_all = Event("BR_INST_RETIRED.ALL_BRANCHES", "BR_INST_RETIRED.ANY")
        br_m_tk = None
        try:
            br_m_tk = Event("BR_MISP_RETIRED.NEAR_TAKEN",
                            "BR_MISP_RETIRED.TAKEN_JCC",
                            "BR_INST_RETIRED.MISPRED_TAKEN")
        except:
            pass
        br_r = d_ratio(br_all, interval_sec)
        ins_r = d_ratio(ins, br_all)
        misp_r = d_ratio(br_m_tk, br_all) if br_m_tk else None
        return MetricGroup("lpm_br_taken", [
            Metric("lpm_br_taken_retired",
                   "The number of taken branches that were retired per second.",
                   br_r, "insn/s"),
            Metric(
                "lpm_br_taken_mispred",
                "The number of retired taken branch instructions that were "
                "mispredicted as a percentage of all taken branches.", misp_r,
                "100%") if misp_r else None,
            Metric(
                "lpm_br_taken_insn_between_branches",
                "The number of instructions divided by the number of taken branches.",
                ins_r, "insn"),
        ])

    def Conditional() -> Optional[MetricGroup]:
        try:
            br_cond = Event("BR_INST_RETIRED.COND",
                            "BR_INST_RETIRED.CONDITIONAL",
                            "BR_INST_RETIRED.TAKEN_JCC")
            br_m_cond = Event("BR_MISP_RETIRED.COND",
                              "BR_MISP_RETIRED.CONDITIONAL",
                              "BR_MISP_RETIRED.TAKEN_JCC")
        except:
            return None

        br_cond_nt = None
        br_m_cond_nt = None
        try:
            br_cond_nt = Event("BR_INST_RETIRED.COND_NTAKEN")
            br_m_cond_nt = Event("BR_MISP_RETIRED.COND_NTAKEN")
        except:
            pass
        br_r = d_ratio(br_cond, interval_sec)
        ins_r = d_ratio(ins, br_cond)
        misp_r = d_ratio(br_m_cond, br_cond)
        taken_metrics = [
            Metric("lpm_br_cond_retired", "Retired conditional branch instructions.",
                   br_r, "insn/s"),
            Metric("lpm_br_cond_insn_between_branches",
                   "The number of instructions divided by the number of conditional "
                   "branches.", ins_r, "insn"),
            Metric("lpm_br_cond_mispred",
                   "Retired conditional branch instructions mispredicted as a "
                   "percentage of all conditional branches.", misp_r, "100%"),
        ]
        if not br_m_cond_nt:
            return MetricGroup("lpm_br_cond", taken_metrics)

        br_r = d_ratio(br_cond_nt, interval_sec)
        ins_r = d_ratio(ins, br_cond_nt)
        misp_r = d_ratio(br_m_cond_nt, br_cond_nt)

        not_taken_metrics = [
            Metric("lpm_br_cond_retired", "Retired conditional not taken branch instructions.",
                   br_r, "insn/s"),
            Metric("lpm_br_cond_insn_between_branches",
                   "The number of instructions divided by the number of not taken conditional "
                   "branches.", ins_r, "insn"),
            Metric("lpm_br_cond_mispred",
                   "Retired not taken conditional branch instructions mispredicted as a "
                   "percentage of all not taken conditional branches.", misp_r, "100%"),
        ]
        return MetricGroup("lpm_br_cond", [
            MetricGroup("lpm_br_cond_nt", not_taken_metrics),
            MetricGroup("lpm_br_cond_tkn", taken_metrics),
        ])

    def Far() -> Optional[MetricGroup]:
        try:
            br_far = Event("BR_INST_RETIRED.FAR_BRANCH")
        except:
            return None

        br_r = d_ratio(br_far, interval_sec)
        ins_r = d_ratio(ins, br_far)
        return MetricGroup("lpm_br_far", [
            Metric("lpm_br_far_retired", "Retired far control transfers per second.",
                   br_r, "insn/s"),
            Metric(
                "lpm_br_far_insn_between_branches",
                "The number of instructions divided by the number of far branches.",
                ins_r, "insn"),
        ])

    return MetricGroup("lpm_br", [Total(), Taken(), Conditional(), Far()],
                       description="breakdown of retired branch instructions")


def IntelCtxSw() -> MetricGroup:
    cs = Event("context\\-switches")
    metrics = [
        Metric("lpm_cs_rate", "Context switches per second",
               d_ratio(cs, interval_sec), "ctxsw/s")
    ]

    ev = Event("instructions")
    metrics.append(Metric("lpm_cs_instr", "Instructions per context switch",
                          d_ratio(ev, cs), "instr/cs"))

    ev = Event("cycles")
    metrics.append(Metric("lpm_cs_cycles", "Cycles per context switch",
                          d_ratio(ev, cs), "cycles/cs"))

    try:
        ev = Event("MEM_INST_RETIRED.ALL_LOADS", "MEM_UOPS_RETIRED.ALL_LOADS")
        metrics.append(Metric("lpm_cs_loads", "Loads per context switch",
                              d_ratio(ev, cs), "loads/cs"))
    except:
        pass

    try:
        ev = Event("MEM_INST_RETIRED.ALL_STORES",
                   "MEM_UOPS_RETIRED.ALL_STORES")
        metrics.append(Metric("lpm_cs_stores", "Stores per context switch",
                              d_ratio(ev, cs), "stores/cs"))
    except:
        pass

    try:
        ev = Event("BR_INST_RETIRED.NEAR_TAKEN", "BR_INST_RETIRED.TAKEN_JCC")
        metrics.append(Metric("lpm_cs_br_taken", "Branches taken per context switch",
                              d_ratio(ev, cs), "br_taken/cs"))
    except:
        pass

    try:
        l2_misses = (Event("L2_RQSTS.DEMAND_DATA_RD_MISS") +
                     Event("L2_RQSTS.RFO_MISS") +
                     Event("L2_RQSTS.CODE_RD_MISS"))
        try:
            l2_misses += Event("L2_RQSTS.HWPF_MISS",
                               "L2_RQSTS.L2_PF_MISS", "L2_RQSTS.PF_MISS")
        except:
            pass

        metrics.append(Metric("lpm_cs_l2_misses", "L2 misses per context switch",
                              d_ratio(l2_misses, cs), "l2_misses/cs"))
    except:
        pass

    return MetricGroup("lpm_cs", metrics,
                       description=("Number of context switches per second, instructions "
                                    "retired & core cycles between context switches"))


def IntelFpu() -> Optional[MetricGroup]:
    cyc = Event("cycles")
    try:
        s_64 = Event("FP_ARITH_INST_RETIRED.SCALAR_SINGLE",
                     "SIMD_INST_RETIRED.SCALAR_SINGLE")
    except:
        return None
    d_64 = Event("FP_ARITH_INST_RETIRED.SCALAR_DOUBLE",
                 "SIMD_INST_RETIRED.SCALAR_DOUBLE")
    s_128 = Event("FP_ARITH_INST_RETIRED.128B_PACKED_SINGLE",
                  "SIMD_INST_RETIRED.PACKED_SINGLE")

    flop = s_64 + d_64 + 4 * s_128

    d_128 = None
    s_256 = None
    d_256 = None
    s_512 = None
    d_512 = None
    try:
        d_128 = Event("FP_ARITH_INST_RETIRED.128B_PACKED_DOUBLE")
        flop += 2 * d_128
        s_256 = Event("FP_ARITH_INST_RETIRED.256B_PACKED_SINGLE")
        flop += 8 * s_256
        d_256 = Event("FP_ARITH_INST_RETIRED.256B_PACKED_DOUBLE")
        flop += 4 * d_256
        s_512 = Event("FP_ARITH_INST_RETIRED.512B_PACKED_SINGLE")
        flop += 16 * s_512
        d_512 = Event("FP_ARITH_INST_RETIRED.512B_PACKED_DOUBLE")
        flop += 8 * d_512
    except:
        pass

    f_assist = Event("ASSISTS.FP", "FP_ASSIST.ANY", "FP_ASSIST.S")
    if f_assist in [
        "ASSISTS.FP",
        "FP_ASSIST.S",
    ]:
        f_assist += "/cmask=1/"

    flop_r = d_ratio(flop, interval_sec)
    flop_c = d_ratio(flop, cyc)
    nmi_constraint = MetricConstraint.GROUPED_EVENTS
    if f_assist.name == "ASSISTS.FP":  # Icelake+
        nmi_constraint = MetricConstraint.NO_GROUP_EVENTS_NMI

    def FpuMetrics(group: str, fl: Optional[Event], mult: int, desc: str) -> Optional[MetricGroup]:
        if not fl:
            return None

        f = fl * mult
        fl_r = d_ratio(f, interval_sec)
        r_s = d_ratio(fl, interval_sec)
        return MetricGroup(group, [
            Metric(f"{group}_of_total", desc + " floating point operations per second",
                   d_ratio(f, flop), "100%"),
            Metric(f"{group}_flops", desc + " floating point operations per second",
                   fl_r, "flops/s"),
            Metric(f"{group}_ops", desc + " operations per second",
                   r_s, "ops/s"),
        ])

    return MetricGroup("lpm_fpu", [
        MetricGroup("lpm_fpu_total", [
            Metric("lpm_fpu_total_flops", "Floating point operations per second",
                   flop_r, "flops/s"),
            Metric("lpm_fpu_total_flopc", "Floating point operations per cycle",
                   flop_c, "flops/cycle", constraint=nmi_constraint),
        ]),
        MetricGroup("lpm_fpu_64", [
            FpuMetrics("lpm_fpu_64_single", s_64, 1, "64-bit single"),
            FpuMetrics("lpm_fpu_64_double", d_64, 1, "64-bit double"),
        ]),
        MetricGroup("lpm_fpu_128", [
            FpuMetrics("lpm_fpu_128_single", s_128,
                       4, "128-bit packed single"),
            FpuMetrics("lpm_fpu_128_double", d_128,
                       2, "128-bit packed double"),
        ]),
        MetricGroup("lpm_fpu_256", [
            FpuMetrics("lpm_fpu_256_single", s_256,
                       8, "128-bit packed single"),
            FpuMetrics("lpm_fpu_256_double", d_256,
                       4, "128-bit packed double"),
        ]),
        MetricGroup("lpm_fpu_512", [
            FpuMetrics("lpm_fpu_512_single", s_512,
                       16, "128-bit packed single"),
            FpuMetrics("lpm_fpu_512_double", d_512,
                       8, "128-bit packed double"),
        ]),
        Metric("lpm_fpu_assists", "FP assists as a percentage of cycles",
               d_ratio(f_assist, cyc), "100%"),
    ])


def IntelIlp() -> MetricGroup:
    tsc = Event("msr/tsc/")
    c0 = Event("msr/mperf/")
    low = tsc - c0
    inst_ret = Event("INST_RETIRED.ANY_P")
    inst_ret_c = [Event(f"{inst_ret.name}/cmask={x}/") for x in range(1, 6)]
    core_cycles = Event("CPU_CLK_UNHALTED.THREAD_P_ANY",
                        "CPU_CLK_UNHALTED.DISTRIBUTED",
                        "cycles")
    ilp = [d_ratio(max(inst_ret_c[x] - inst_ret_c[x + 1], 0), core_cycles)
           for x in range(0, 4)]
    ilp.append(d_ratio(inst_ret_c[4], core_cycles))
    ilp0 = 1
    for x in ilp:
        ilp0 -= x
    return MetricGroup("lpm_ilp", [
        Metric("lpm_ilp_idle", "Lower power cycles as a percentage of all cycles",
               d_ratio(low, tsc), "100%"),
        Metric("lpm_ilp_inst_ret_0",
               "Instructions retired in 0 cycles as a percentage of all cycles",
               ilp0, "100%"),
        Metric("lpm_ilp_inst_ret_1",
               "Instructions retired in 1 cycles as a percentage of all cycles",
               ilp[0], "100%"),
        Metric("lpm_ilp_inst_ret_2",
               "Instructions retired in 2 cycles as a percentage of all cycles",
               ilp[1], "100%"),
        Metric("lpm_ilp_inst_ret_3",
               "Instructions retired in 3 cycles as a percentage of all cycles",
               ilp[2], "100%"),
        Metric("lpm_ilp_inst_ret_4",
               "Instructions retired in 4 cycles as a percentage of all cycles",
               ilp[3], "100%"),
        Metric("lpm_ilp_inst_ret_5",
               "Instructions retired in 5 or more cycles as a percentage of all cycles",
               ilp[4], "100%"),
    ])


def IntelIotlb() -> Optional[MetricGroup]:
  try:
    total_hit = (
        Event("UNC_IIO_IOMMU0.4K_HITS")
        + Event("UNC_IIO_IOMMU0.2M_HITS")
        + Event("UNC_IIO_IOMMU0.1G_HITS")
    )
    total_miss = Event("UNC_IIO_IOMMU0.MISSES")
  except:
    return None

  miss_rate = d_ratio(total_miss, total_miss + total_hit)
  metrics = [
      Metric("iotlb_total_hit", "IOTLB total hit", total_hit, "hits"),
      Metric("iotlb_total_miss", "IOTLB total miss", total_miss, "misses"),
      Metric("iotlb_miss_rate", "IOTLB miss rate", miss_rate, "100%"),
  ]

  try:
    interrupt_cache_hit = Event("UNC_IIO_IOMMU3.INT_CACHE_HITS")
    interrupt_cache_lookup = Event("UNC_IIO_IOMMU3.INT_CACHE_LOOKUPS")
    interrupt_cache_miss = max(interrupt_cache_lookup - interrupt_cache_hit, 0)
    interrupt_cache_miss_rate = d_ratio(
        interrupt_cache_miss, interrupt_cache_miss + interrupt_cache_hit
    )
    metrics += [
        Metric(
            "iotlb_interrupt_cache_hit",
            "IOTLB interrupt cache hit",
            interrupt_cache_hit,
            "hits",
        ),
        Metric(
            "iotlb_interrupt_cache_miss",
            "IOTLB interrupt cache miss",
            interrupt_cache_miss,
            "misses",
        ),
        Metric(
            "iotlb_interrupt_cache_lookup",
            "IOTLB interrupt cache lookup",
            interrupt_cache_lookup,
            "lookups",
        ),
        Metric(
            "iotlb_interrupt_cache_miss_rate",
            "IOTLB interrupt cache miss rate",
            interrupt_cache_miss_rate,
            "100%",
        ),
    ]
  except:
    pass

  return MetricGroup(
      "iotlb",
      metrics,
      description="IOMMU TLB metrics",
  )


def IntelL2() -> Optional[MetricGroup]:
    try:
        DC_HIT = Event("L2_RQSTS.DEMAND_DATA_RD_HIT")
    except:
        return None
    try:
        DC_MISS = Event("L2_RQSTS.DEMAND_DATA_RD_MISS")
        l2_dmnd_miss = DC_MISS
        l2_dmnd_rd_all = DC_MISS + DC_HIT
    except:
        DC_ALL = Event("L2_RQSTS.ALL_DEMAND_DATA_RD")
        l2_dmnd_miss = DC_ALL - DC_HIT
        l2_dmnd_rd_all = DC_ALL
    l2_dmnd_mrate = d_ratio(l2_dmnd_miss, interval_sec)
    l2_dmnd_rrate = d_ratio(l2_dmnd_rd_all, interval_sec)

    DC_PFH = None
    DC_PFM = None
    l2_pf_all = None
    l2_pf_mrate = None
    l2_pf_rrate = None
    try:
        DC_PFH = Event("L2_RQSTS.PF_HIT")
        DC_PFM = Event("L2_RQSTS.PF_MISS")
        l2_pf_all = DC_PFH + DC_PFM
        l2_pf_mrate = d_ratio(DC_PFM, interval_sec)
        l2_pf_rrate = d_ratio(l2_pf_all, interval_sec)
    except:
        pass

    DC_RFOH = None
    DC_RFOM = None
    l2_rfo_all = None
    l2_rfo_mrate = None
    l2_rfo_rrate = None
    try:
        DC_RFOH = Event("L2_RQSTS.RFO_HIT")
        DC_RFOM = Event("L2_RQSTS.RFO_MISS")
        l2_rfo_all = DC_RFOH + DC_RFOM
        l2_rfo_mrate = d_ratio(DC_RFOM, interval_sec)
        l2_rfo_rrate = d_ratio(l2_rfo_all, interval_sec)
    except:
        pass

    DC_CH = None
    try:
        DC_CH = Event("L2_RQSTS.CODE_RD_HIT")
    except:
        pass
    DC_CM = Event("L2_RQSTS.CODE_RD_MISS")
    DC_IN = Event("L2_LINES_IN.ALL")
    DC_OUT_NS = None
    DC_OUT_S = None
    l2_lines_out = None
    l2_out_rate = None
    wbn = None
    isd = None
    try:
        DC_OUT_NS = Event("L2_LINES_OUT.NON_SILENT",
                          "L2_LINES_OUT.DEMAND_DIRTY",
                          "L2_LINES_IN.S")
        DC_OUT_S = Event("L2_LINES_OUT.SILENT",
                         "L2_LINES_OUT.DEMAND_CLEAN",
                         "L2_LINES_IN.I")
        if DC_OUT_S.name == "L2_LINES_OUT.SILENT" and (
                args.model.startswith("skylake") or
                args.model == "cascadelakex"):
            DC_OUT_S.name = "L2_LINES_OUT.SILENT/any/"
        # bring is back to per-CPU
        l2_s = Select(DC_OUT_S / 2, Literal("#smt_on"), DC_OUT_S)
        l2_ns = DC_OUT_NS
        l2_lines_out = l2_s + l2_ns
        l2_out_rate = d_ratio(l2_lines_out, interval_sec)
        nlr = max(l2_ns - DC_WB_U - DC_WB_D, 0)
        wbn = d_ratio(nlr, interval_sec)
        isd = d_ratio(l2_s, interval_sec)
    except:
        pass
    DC_OUT_U = None
    l2_pf_useless = None
    l2_useless_rate = None
    try:
        DC_OUT_U = Event("L2_LINES_OUT.USELESS_HWPF")
        l2_pf_useless = DC_OUT_U
        l2_useless_rate = d_ratio(l2_pf_useless, interval_sec)
    except:
        pass
    DC_WB_U = None
    DC_WB_D = None
    wbu = None
    wbd = None
    try:
        DC_WB_U = Event("IDI_MISC.WB_UPGRADE")
        DC_WB_D = Event("IDI_MISC.WB_DOWNGRADE")
        wbu = d_ratio(DC_WB_U, interval_sec)
        wbd = d_ratio(DC_WB_D, interval_sec)
    except:
        pass

    l2_lines_in = DC_IN
    l2_code_all = (DC_CH + DC_CM) if DC_CH else None
    l2_code_rate = d_ratio(l2_code_all, interval_sec) if DC_CH else None
    l2_code_miss_rate = d_ratio(DC_CM, interval_sec)
    l2_in_rate = d_ratio(l2_lines_in, interval_sec)

    return MetricGroup("lpm_l2", [
        MetricGroup("lpm_l2_totals", [
            Metric("lpm_l2_totals_in", "L2 cache total in per second",
                   l2_in_rate, "In/s"),
            Metric("lpm_l2_totals_out", "L2 cache total out per second",
                   l2_out_rate, "Out/s") if l2_out_rate else None,
        ]),
        MetricGroup("lpm_l2_rd", [
            Metric("lpm_l2_rd_hits", "L2 cache data read hits",
                   d_ratio(DC_HIT, l2_dmnd_rd_all), "100%"),
            Metric("lpm_l2_rd_hits", "L2 cache data read hits",
                   d_ratio(l2_dmnd_miss, l2_dmnd_rd_all), "100%"),
            Metric("lpm_l2_rd_requests", "L2 cache data read requests per second",
                   l2_dmnd_rrate, "requests/s"),
            Metric("lpm_l2_rd_misses", "L2 cache data read misses per second",
                   l2_dmnd_mrate, "misses/s"),
        ]),
        MetricGroup("lpm_l2_hwpf", [
            Metric("lpm_l2_hwpf_hits", "L2 cache hardware prefetcher hits",
                   d_ratio(DC_PFH, l2_pf_all), "100%"),
            Metric("lpm_l2_hwpf_misses", "L2 cache hardware prefetcher misses",
                   d_ratio(DC_PFM, l2_pf_all), "100%"),
            Metric("lpm_l2_hwpf_useless", "L2 cache hardware prefetcher useless prefetches per second",
                   l2_useless_rate, "100%") if l2_useless_rate else None,
            Metric("lpm_l2_hwpf_requests", "L2 cache hardware prefetcher requests per second",
                   l2_pf_rrate, "100%"),
            Metric("lpm_l2_hwpf_misses", "L2 cache hardware prefetcher misses per second",
                   l2_pf_mrate, "100%"),
        ]) if DC_PFH else None,
        MetricGroup("lpm_l2_rfo", [
            Metric("lpm_l2_rfo_hits", "L2 cache request for ownership (RFO) hits",
                   d_ratio(DC_RFOH, l2_rfo_all), "100%"),
            Metric("lpm_l2_rfo_misses", "L2 cache request for ownership (RFO) misses",
                   d_ratio(DC_RFOM, l2_rfo_all), "100%"),
            Metric("lpm_l2_rfo_requests", "L2 cache request for ownership (RFO) requests per second",
                   l2_rfo_rrate, "requests/s"),
            Metric("lpm_l2_rfo_misses", "L2 cache request for ownership (RFO) misses per second",
                   l2_rfo_mrate, "misses/s"),
        ]) if DC_RFOH else None,
        MetricGroup("lpm_l2_code", [
            Metric("lpm_l2_code_hits", "L2 cache code hits",
                   d_ratio(DC_CH, l2_code_all), "100%") if DC_CH else None,
            Metric("lpm_l2_code_misses", "L2 cache code misses",
                   d_ratio(DC_CM, l2_code_all), "100%") if DC_CH else None,
            Metric("lpm_l2_code_requests", "L2 cache code requests per second",
                   l2_code_rate, "requests/s") if DC_CH else None,
            Metric("lpm_l2_code_misses", "L2 cache code misses per second",
                   l2_code_miss_rate, "misses/s"),
        ]),
        MetricGroup("lpm_l2_evict", [
            MetricGroup("lpm_l2_evict_mef_lines", [
                Metric("lpm_l2_evict_mef_lines_l3_hot_lru", "L2 evictions M/E/F lines L3 hot LRU per second",
                       wbu, "HotLRU/s") if wbu else None,
                Metric("lpm_l2_evict_mef_lines_l3_norm_lru", "L2 evictions M/E/F lines L3 normal LRU per second",
                       wbn, "NormLRU/s") if wbn else None,
                Metric("lpm_l2_evict_mef_lines_dropped", "L2 evictions M/E/F lines dropped per second",
                       wbd, "dropped/s") if wbd else None,
                Metric("lpm_l2_evict_is_lines_dropped", "L2 evictions I/S lines dropped per second",
                       isd, "dropped/s") if isd else None,
            ]),
        ]),
    ], description="L2 data cache analysis")


def IntelMissLat() -> Optional[MetricGroup]:
    try:
        ticks = Event("UNC_CHA_CLOCKTICKS", "UNC_C_CLOCKTICKS")
        data_rd_loc_occ = Event("UNC_CHA_TOR_OCCUPANCY.IA_MISS_DRD_LOCAL",
                                "UNC_CHA_TOR_OCCUPANCY.IA_MISS",
                                "UNC_C_TOR_OCCUPANCY.MISS_LOCAL_OPCODE",
                                "UNC_C_TOR_OCCUPANCY.MISS_OPCODE")
        data_rd_loc_ins = Event("UNC_CHA_TOR_INSERTS.IA_MISS_DRD_LOCAL",
                                "UNC_CHA_TOR_INSERTS.IA_MISS",
                                "UNC_C_TOR_INSERTS.MISS_LOCAL_OPCODE",
                                "UNC_C_TOR_INSERTS.MISS_OPCODE")
        data_rd_rem_occ = Event("UNC_CHA_TOR_OCCUPANCY.IA_MISS_DRD_REMOTE",
                                "UNC_CHA_TOR_OCCUPANCY.IA_MISS",
                                "UNC_C_TOR_OCCUPANCY.MISS_REMOTE_OPCODE",
                                "UNC_C_TOR_OCCUPANCY.NID_MISS_OPCODE")
        data_rd_rem_ins = Event("UNC_CHA_TOR_INSERTS.IA_MISS_DRD_REMOTE",
                                "UNC_CHA_TOR_INSERTS.IA_MISS",
                                "UNC_C_TOR_INSERTS.MISS_REMOTE_OPCODE",
                                "UNC_C_TOR_INSERTS.NID_MISS_OPCODE")
    except:
        return None

    if (data_rd_loc_occ.name == "UNC_C_TOR_OCCUPANCY.MISS_LOCAL_OPCODE" or
            data_rd_loc_occ.name == "UNC_C_TOR_OCCUPANCY.MISS_OPCODE"):
        data_rd = 0x182
        for e in [data_rd_loc_occ, data_rd_loc_ins, data_rd_rem_occ, data_rd_rem_ins]:
            e.name += f"/filter_opc={hex(data_rd)}/"
    elif data_rd_loc_occ.name == "UNC_CHA_TOR_OCCUPANCY.IA_MISS":
        # Demand Data Read - Full cache-line read requests from core for
        # lines to be cached in S or E, typically for data
        demand_data_rd = 0x202
        #  LLC Prefetch Data - Uncore will first look up the line in the
        #  LLC; for a cache hit, the LRU will be updated, on a miss, the
        #  DRd will be initiated
        llc_prefetch_data = 0x25a
        local_filter = (f"/filter_opc0={hex(demand_data_rd)},"
                        f"filter_opc1={hex(llc_prefetch_data)},"
                        "filter_loc,filter_nm,filter_not_nm/")
        remote_filter = (f"/filter_opc0={hex(demand_data_rd)},"
                         f"filter_opc1={hex(llc_prefetch_data)},"
                         "filter_rem,filter_nm,filter_not_nm/")
        for e in [data_rd_loc_occ, data_rd_loc_ins]:
            e.name += local_filter
        for e in [data_rd_rem_occ, data_rd_rem_ins]:
            e.name += remote_filter
    else:
        assert data_rd_loc_occ.name == "UNC_CHA_TOR_OCCUPANCY.IA_MISS_DRD_LOCAL", data_rd_loc_occ

    ticks_per_cha = ticks / aggr_nr(data_rd_loc_ins)
    loc_lat = interval_sec * 1e9 * data_rd_loc_occ / \
        (ticks_per_cha * data_rd_loc_ins)
    ticks_per_cha = ticks / aggr_nr(data_rd_rem_ins)
    rem_lat = interval_sec * 1e9 * data_rd_rem_occ / \
        (ticks_per_cha * data_rd_rem_ins)
    return MetricGroup("lpm_miss_lat", [
        Metric("lpm_miss_lat_loc", "Local to a socket miss latency in nanoseconds",
               loc_lat, "ns"),
        Metric("lpm_miss_lat_rem", "Remote to a socket miss latency in nanoseconds",
               rem_lat, "ns"),
    ])


def IntelMlp() -> Optional[Metric]:
    try:
        l1d = Event("L1D_PEND_MISS.PENDING")
        l1dc = Event("L1D_PEND_MISS.PENDING_CYCLES")
    except:
        return None

    l1dc = Select(l1dc / 2, Literal("#smt_on"), l1dc)
    ml = d_ratio(l1d, l1dc)
    return Metric("lpm_mlp",
                  "Miss level parallelism - number of outstanding load misses per cycle (higher is better)",
                  ml, "load_miss_pending/cycle")


def IntelPorts() -> Optional[MetricGroup]:
    pipeline_events = json.load(
        open(f"{_args.events_path}/x86/{_args.model}/pipeline.json"))

    core_cycles = Event("CPU_CLK_UNHALTED.THREAD_P_ANY",
                        "CPU_CLK_UNHALTED.DISTRIBUTED",
                        "cycles")
    # Number of CPU cycles scaled for SMT.
    smt_cycles = Select(core_cycles / 2, Literal("#smt_on"), core_cycles)

    metrics = []
    for x in pipeline_events:
        if "EventName" in x and re.search("^UOPS_DISPATCHED.PORT", x["EventName"]):
            name = x["EventName"]
            port = re.search(r"(PORT_[0-9].*)", name).group(0).lower()
            if name.endswith("_CORE"):
                cyc = core_cycles
            else:
                cyc = smt_cycles
            metrics.append(Metric(f"lpm_{port}", f"{port} utilization (higher is better)",
                                  d_ratio(Event(name), cyc), "100%"))
    if len(metrics) == 0:
        return None

    return MetricGroup("lpm_ports", metrics, "functional unit (port) utilization -- "
                       "fraction of cycles each port is utilized (higher is better)")


def IntelSwpf() -> Optional[MetricGroup]:
    ins = Event("instructions")
    try:
        s_ld = Event("MEM_INST_RETIRED.ALL_LOADS",
                     "MEM_UOPS_RETIRED.ALL_LOADS")
        s_nta = Event("SW_PREFETCH_ACCESS.NTA")
        s_t0 = Event("SW_PREFETCH_ACCESS.T0")
        s_t1 = Event("SW_PREFETCH_ACCESS.T1_T2")
        s_w = Event("SW_PREFETCH_ACCESS.PREFETCHW")
    except:
        return None

    all_sw = s_nta + s_t0 + s_t1 + s_w
    swp_r = d_ratio(all_sw, interval_sec)
    ins_r = d_ratio(ins, all_sw)
    ld_r = d_ratio(s_ld, all_sw)

    return MetricGroup("lpm_swpf", [
        MetricGroup("lpm_swpf_totals", [
            Metric("lpm_swpf_totals_exec", "Software prefetch instructions per second",
                   swp_r, "swpf/s"),
            Metric("lpm_swpf_totals_insn_per_pf",
                   "Average number of instructions between software prefetches",
                   ins_r, "insn/swpf"),
            Metric("lpm_swpf_totals_loads_per_pf",
                   "Average number of loads between software prefetches",
                   ld_r, "loads/swpf"),
        ]),
        MetricGroup("lpm_swpf_bkdwn", [
            MetricGroup("lpm_swpf_bkdwn_nta", [
                Metric("lpm_swpf_bkdwn_nta_per_swpf",
                       "Software prefetch NTA instructions as a percent of all prefetch instructions",
                       d_ratio(s_nta, all_sw), "100%"),
                Metric("lpm_swpf_bkdwn_nta_rate",
                       "Software prefetch NTA instructions per second",
                       d_ratio(s_nta, interval_sec), "insn/s"),
            ]),
            MetricGroup("lpm_swpf_bkdwn_t0", [
                Metric("lpm_swpf_bkdwn_t0_per_swpf",
                       "Software prefetch T0 instructions as a percent of all prefetch instructions",
                       d_ratio(s_t0, all_sw), "100%"),
                Metric("lpm_swpf_bkdwn_t0_rate",
                       "Software prefetch T0 instructions per second",
                       d_ratio(s_t0, interval_sec), "insn/s"),
            ]),
            MetricGroup("lpm_swpf_bkdwn_t1_t2", [
                Metric("lpm_swpf_bkdwn_t1_t2_per_swpf",
                       "Software prefetch T1 or T2 instructions as a percent of all prefetch instructions",
                       d_ratio(s_t1, all_sw), "100%"),
                Metric("lpm_swpf_bkdwn_t1_t2_rate",
                       "Software prefetch T1 or T2 instructions per second",
                       d_ratio(s_t1, interval_sec), "insn/s"),
            ]),
            MetricGroup("lpm_swpf_bkdwn_w", [
                Metric("lpm_swpf_bkdwn_w_per_swpf",
                       "Software prefetch W instructions as a percent of all prefetch instructions",
                       d_ratio(s_w, all_sw), "100%"),
                Metric("lpm_swpf_bkdwn_w_rate",
                       "Software prefetch W instructions per second",
                       d_ratio(s_w, interval_sec), "insn/s"),
            ]),
        ]),
    ], description="Software prefetch instruction breakdown")


def IntelLdSt() -> Optional[MetricGroup]:
    if _args.model in [
        "bonnell",
        "nehalemep",
        "nehalemex",
        "westmereep-dp",
        "westmereep-sp",
        "westmereex",
    ]:
        return None
    LDST_LD = Event("MEM_INST_RETIRED.ALL_LOADS", "MEM_UOPS_RETIRED.ALL_LOADS")
    LDST_ST = Event("MEM_INST_RETIRED.ALL_STORES",
                    "MEM_UOPS_RETIRED.ALL_STORES")
    LDST_LDC1 = Event(f"{LDST_LD.name}/cmask=1/")
    LDST_STC1 = Event(f"{LDST_ST.name}/cmask=1/")
    LDST_LDC2 = Event(f"{LDST_LD.name}/cmask=2/")
    LDST_STC2 = Event(f"{LDST_ST.name}/cmask=2/")
    LDST_LDC3 = Event(f"{LDST_LD.name}/cmask=3/")
    LDST_STC3 = Event(f"{LDST_ST.name}/cmask=3/")
    ins = Event("instructions")
    LDST_CYC = Event("CPU_CLK_UNHALTED.THREAD",
                     "CPU_CLK_UNHALTED.CORE_P",
                     "CPU_CLK_UNHALTED.THREAD_P")
    LDST_PRE = None
    try:
        LDST_PRE = Event("LOAD_HIT_PREFETCH.SWPF", "LOAD_HIT_PRE.SW_PF")
    except:
        pass
    LDST_AT = None
    try:
        LDST_AT = Event("MEM_INST_RETIRED.LOCK_LOADS")
    except:
        pass
    cyc = LDST_CYC

    ld_rate = d_ratio(LDST_LD, interval_sec)
    st_rate = d_ratio(LDST_ST, interval_sec)
    pf_rate = d_ratio(LDST_PRE, interval_sec) if LDST_PRE else None
    at_rate = d_ratio(LDST_AT, interval_sec) if LDST_AT else None

    ldst_ret_constraint = MetricConstraint.GROUPED_EVENTS
    if LDST_LD.name == "MEM_UOPS_RETIRED.ALL_LOADS":
        ldst_ret_constraint = MetricConstraint.NO_GROUP_EVENTS_NMI

    return MetricGroup("lpm_ldst", [
        MetricGroup("lpm_ldst_total", [
            Metric("lpm_ldst_total_loads", "Load/store instructions total loads",
                   ld_rate, "loads"),
            Metric("lpm_ldst_total_stores", "Load/store instructions total stores",
                   st_rate, "stores"),
        ]),
        MetricGroup("lpm_ldst_prcnt", [
            Metric("lpm_ldst_prcnt_loads", "Percent of all instructions that are loads",
                   d_ratio(LDST_LD, ins), "100%"),
            Metric("lpm_ldst_prcnt_stores", "Percent of all instructions that are stores",
                   d_ratio(LDST_ST, ins), "100%"),
        ]),
        MetricGroup("lpm_ldst_ret_lds", [
            Metric("lpm_ldst_ret_lds_1", "Retired loads in 1 cycle",
                   d_ratio(max(LDST_LDC1 - LDST_LDC2, 0), cyc), "100%",
                   constraint=ldst_ret_constraint),
            Metric("lpm_ldst_ret_lds_2", "Retired loads in 2 cycles",
                   d_ratio(max(LDST_LDC2 - LDST_LDC3, 0), cyc), "100%",
                   constraint=ldst_ret_constraint),
            Metric("lpm_ldst_ret_lds_3", "Retired loads in 3 or more cycles",
                   d_ratio(LDST_LDC3, cyc), "100%"),
        ]),
        MetricGroup("lpm_ldst_ret_sts", [
            Metric("lpm_ldst_ret_sts_1", "Retired stores in 1 cycle",
                   d_ratio(max(LDST_STC1 - LDST_STC2, 0), cyc), "100%",
                   constraint=ldst_ret_constraint),
            Metric("lpm_ldst_ret_sts_2", "Retired stores in 2 cycles",
                   d_ratio(max(LDST_STC2 - LDST_STC3, 0), cyc), "100%",
                   constraint=ldst_ret_constraint),
            Metric("lpm_ldst_ret_sts_3", "Retired stores in 3 more cycles",
                   d_ratio(LDST_STC3, cyc), "100%"),
        ]),
        Metric("lpm_ldst_ld_hit_swpf", "Load hit software prefetches per second",
               pf_rate, "swpf/s") if pf_rate else None,
        Metric("lpm_ldst_atomic_lds", "Atomic loads per second",
               at_rate, "loads/s") if at_rate else None,
    ], description="Breakdown of load/store instructions")


def UncoreCState() -> Optional[MetricGroup]:
    try:
        pcu_ticks = Event("UNC_P_CLOCKTICKS")
        c0 = Event("UNC_P_POWER_STATE_OCCUPANCY.CORES_C0")
        c3 = Event("UNC_P_POWER_STATE_OCCUPANCY.CORES_C3")
        c6 = Event("UNC_P_POWER_STATE_OCCUPANCY.CORES_C6")
    except:
        return None

    num_cores = Literal("#num_cores") / Literal("#num_packages")

    max_cycles = pcu_ticks * num_cores
    total_cycles = c0 + c3 + c6

    # remove fused-off cores which show up in C6/C7.
    c6 = Select(max(c6 - (total_cycles - max_cycles), 0),
                total_cycles > max_cycles,
                c6)

    return MetricGroup("lpm_cstate", [
        Metric("lpm_cstate_c0", "C-State cores in C0/C1",
               d_ratio(c0, pcu_ticks), "cores"),
        Metric("lpm_cstate_c3", "C-State cores in C3",
               d_ratio(c3, pcu_ticks), "cores"),
        Metric("lpm_cstate_c6", "C-State cores in C6/C7",
               d_ratio(c6, pcu_ticks), "cores"),
    ])


def UncoreDir() -> Optional[MetricGroup]:
    try:
        m2m_upd = Event("UNC_M2M_DIRECTORY_UPDATE.ANY")
        m2m_hits = Event("UNC_M2M_DIRECTORY_HIT.DIRTY_I")
        # Turn the umask into a ANY rather than DIRTY_I filter.
        m2m_hits.name += "/umask=0xFF,name=UNC_M2M_DIRECTORY_HIT.ANY/"
        m2m_miss = Event("UNC_M2M_DIRECTORY_MISS.DIRTY_I")
        # Turn the umask into a ANY rather than DIRTY_I filter.
        m2m_miss.name += "/umask=0xFF,name=UNC_M2M_DIRECTORY_MISS.ANY/"
        cha_upd = Event("UNC_CHA_DIR_UPDATE.HA")
        # Turn the umask into a ANY rather than HA filter.
        cha_upd.name += "/umask=3,name=UNC_CHA_DIR_UPDATE.ANY/"
    except:
        return None

    m2m_total = m2m_hits + m2m_miss
    upd = m2m_upd + cha_upd  # in cache lines
    upd_r = upd / interval_sec
    look_r = m2m_total / interval_sec

    scale = 64 / 1_000_000  # Cache lines to MB
    return MetricGroup("lpm_dir", [
        Metric("lpm_dir_lookup_rate", "",
               d_ratio(m2m_total, interval_sec), "requests/s"),
        Metric("lpm_dir_lookup_hits", "",
               d_ratio(m2m_hits, m2m_total), "100%"),
        Metric("lpm_dir_lookup_misses", "",
               d_ratio(m2m_miss, m2m_total), "100%"),
        Metric("lpm_dir_update_requests", "",
               d_ratio(m2m_upd + cha_upd, interval_sec), "requests/s"),
        Metric("lpm_dir_update_bw", "",
               d_ratio(m2m_upd + cha_upd, interval_sec), f"{scale}MB/s"),
    ])


def UncoreMem() -> Optional[MetricGroup]:
    try:
        loc_rds = Event("UNC_CHA_REQUESTS.READS_LOCAL",
                        "UNC_H_REQUESTS.READS_LOCAL")
        rem_rds = Event("UNC_CHA_REQUESTS.READS_REMOTE",
                        "UNC_H_REQUESTS.READS_REMOTE")
        loc_wrs = Event("UNC_CHA_REQUESTS.WRITES_LOCAL",
                        "UNC_H_REQUESTS.WRITES_LOCAL")
        rem_wrs = Event("UNC_CHA_REQUESTS.WRITES_REMOTE",
                        "UNC_H_REQUESTS.WRITES_REMOTE")
    except:
        return None

    scale = 64 / 1_000_000
    return MetricGroup("lpm_mem", [
        MetricGroup("lpm_mem_local", [
            Metric("lpm_mem_local_read", "Local memory read bandwidth not including directory updates",
                   d_ratio(loc_rds, interval_sec), f"{scale}MB/s"),
            Metric("lpm_mem_local_write", "Local memory write bandwidth not including directory updates",
                   d_ratio(loc_wrs, interval_sec), f"{scale}MB/s"),
        ]),
        MetricGroup("lpm_mem_remote", [
            Metric("lpm_mem_remote_read", "Remote memory read bandwidth not including directory updates",
                   d_ratio(rem_rds, interval_sec), f"{scale}MB/s"),
            Metric("lpm_mem_remote_write", "Remote memory write bandwidth not including directory updates",
                   d_ratio(rem_wrs, interval_sec), f"{scale}MB/s"),
        ]),
    ], description="Memory Bandwidth breakdown local vs. remote (remote requests in). directory updates not included")


def UncoreMemBw() -> Optional[MetricGroup]:
    mem_events = []
    try:
        mem_events = json.load(open(f"{os.path.dirname(os.path.realpath(__file__))}"
                                    f"/arch/x86/{args.model}/uncore-memory.json"))
    except:
        pass

    ddr_rds = 0
    ddr_wrs = 0
    ddr_total = 0
    for x in mem_events:
        if "EventName" in x:
            name = x["EventName"]
            if re.search("^UNC_MC[0-9]+_RDCAS_COUNT_FREERUN", name):
                ddr_rds += Event(name)
            elif re.search("^UNC_MC[0-9]+_WRCAS_COUNT_FREERUN", name):
                ddr_wrs += Event(name)
            # elif re.search("^UNC_MC[0-9]+_TOTAL_REQCOUNT_FREERUN", name):
            #  ddr_total += Event(name)

    if ddr_rds == 0:
        try:
            ddr_rds = Event("UNC_M_CAS_COUNT.RD")
            ddr_wrs = Event("UNC_M_CAS_COUNT.WR")
        except:
            return None

    ddr_total = ddr_rds + ddr_wrs

    pmm_rds = 0
    pmm_wrs = 0
    try:
        pmm_rds = Event("UNC_M_PMM_RPQ_INSERTS")
        pmm_wrs = Event("UNC_M_PMM_WPQ_INSERTS")
    except:
        pass

    pmm_total = pmm_rds + pmm_wrs

    scale = 64 / 1_000_000
    return MetricGroup("lpm_mem_bw", [
        MetricGroup("lpm_mem_bw_ddr", [
            Metric("lpm_mem_bw_ddr_read", "DDR memory read bandwidth",
                   d_ratio(ddr_rds, interval_sec), f"{scale}MB/s"),
            Metric("lpm_mem_bw_ddr_write", "DDR memory write bandwidth",
                   d_ratio(ddr_wrs, interval_sec), f"{scale}MB/s"),
            Metric("lpm_mem_bw_ddr_total", "DDR memory write bandwidth",
                   d_ratio(ddr_total, interval_sec), f"{scale}MB/s"),
        ], description="DDR Memory Bandwidth"),
        MetricGroup("lpm_mem_bw_pmm", [
            Metric("lpm_mem_bw_pmm_read", "PMM memory read bandwidth",
                   d_ratio(pmm_rds, interval_sec), f"{scale}MB/s"),
            Metric("lpm_mem_bw_pmm_write", "PMM memory write bandwidth",
                   d_ratio(pmm_wrs, interval_sec), f"{scale}MB/s"),
            Metric("lpm_mem_bw_pmm_total", "PMM memory write bandwidth",
                   d_ratio(pmm_total, interval_sec), f"{scale}MB/s"),
        ], description="PMM Memory Bandwidth") if pmm_rds != 0 else None,
    ], description="Memory Bandwidth")


def UncoreMemSat() -> Optional[Metric]:
    try:
        clocks = Event("UNC_CHA_CLOCKTICKS", "UNC_C_CLOCKTICKS")
        sat = Event("UNC_CHA_DISTRESS_ASSERTED.VERT", "UNC_CHA_FAST_ASSERTED.VERT",
                    "UNC_C_FAST_ASSERTED")
    except:
        return None

    desc = ("Mesh Bandwidth saturation (% CBOX cycles with FAST signal asserted, "
            "include QPI bandwidth saturation), lower is better")
    if "UNC_CHA_" in sat.name:
        desc = ("Mesh Bandwidth saturation (% CHA cycles with FAST signal asserted, "
                "include UPI bandwidth saturation), lower is better")
    return Metric("lpm_mem_sat", desc, d_ratio(sat, clocks), "100%")


def UncoreUpiBw() -> Optional[MetricGroup]:
    try:
        upi_rds = Event("UNC_UPI_RxL_FLITS.ALL_DATA")
        upi_wrs = Event("UNC_UPI_TxL_FLITS.ALL_DATA")
    except:
        return None

    upi_total = upi_rds + upi_wrs

    # From "Uncore Performance Monitoring": When measuring the amount of
    # bandwidth consumed by transmission of the data (i.e. NOT including
    # the header), it should be .ALL_DATA / 9 * 64B.
    scale = (64 / 9) / 1_000_000
    return MetricGroup("lpm_upi_bw", [
        Metric("lpm_upi_bw_read", "UPI read bandwidth",
               d_ratio(upi_rds, interval_sec), f"{scale}MB/s"),
        Metric("lpm_upi_bw_write", "DDR memory write bandwidth",
               d_ratio(upi_wrs, interval_sec), f"{scale}MB/s"),
    ], description="UPI Bandwidth")


def main() -> None:
    global _args

    def dir_path(path: str) -> str:
        """Validate path is a directory for argparse."""
        if os.path.isdir(path):
            return path
        raise argparse.ArgumentTypeError(
            f'\'{path}\' is not a valid directory')

    parser = argparse.ArgumentParser(description="Intel perf json generator")
    parser.add_argument(
        "-metricgroups", help="Generate metricgroups data", action='store_true')
    parser.add_argument("model", help="e.g. skylakex")
    parser.add_argument(
        'events_path',
        type=dir_path,
        help='Root of tree containing architecture directories containing json files'
    )
    _args = parser.parse_args()

    directory = f"{_args.events_path}/x86/{_args.model}/"
    LoadEvents(directory)

    all_metrics = MetricGroup("", [
        Cycles(),
        Idle(),
        Rapl(),
        Smi(),
        Tsx(),
        IntelBr(),
        IntelCtxSw(),
        IntelFpu(),
        IntelIlp(),
        IntelIotlb(),
        IntelL2(),
        IntelLdSt(),
        IntelMissLat(),
        IntelMlp(),
        IntelPorts(),
        IntelSwpf(),
        UncoreCState(),
        UncoreDir(),
        UncoreMem(),
        UncoreMemBw(),
        UncoreMemSat(),
        UncoreUpiBw(),
    ])

    if _args.metricgroups:
        print(JsonEncodeMetricGroupDescriptions(all_metrics))
    else:
        print(JsonEncodeMetric(all_metrics))


if __name__ == '__main__':
    main()
